import datetime
import json
import json

isEncyptOut = False
isEncyptIn = True

MAX_SCRIPS = 100
topic_list = {}
counter = 0
FieldTypes = {
    'FLOAT32': 1,
    'LONG': 2,
    'DATE': 3,
    'STRING': 4
}
TRASH_VAL = -2147483648
STRING_INDEX = {
    'NAME': 51,
    'SYMBOL': 52,
    'EXCHG': 53,
    'TSYMBOL': 54
}
DEPTH_INDEX = {
    "MULTIPLIER": 32,
    "PRECISION": 33
}
BinRespTypes = {
    "CONNECTION_TYPE": 1,
    "THROTTLING_TYPE": 2,
    "ACK_TYPE": 3,
    "SUBSCRIBE_TYPE": 4,
    "UNSUBSCRIBE_TYPE": 5,
    "DATA_TYPE": 6,
    "CHPAUSE_TYPE": 7,
    "CHRESUME_TYPE": 8,
    "SNAPSHOT": 9,
    "OPC_SUBSCRIBE": 10
}
# ws = None
BinRespStat = {
    "OK": "K",
    "NOT_OK": "N"
}
ResponseTypes = {
    "SNAP": 83,
    "UPDATE": 85
}
STAT = {
    "OK": "Ok",
    "NOT_OK": "NotOk"
}
RespTypeValues = {
    "CONN": "cn",
    "SUBS": "sub",
    "UNSUBS": "unsub",
    "SNAP": "snap",
    "CHANNELR": "cr",
    "CHANNELP": "cp",
    "OPC": "opc"
}
RespCodes = {
    'SUCCESS': 200,
    'CONNECTION_FAILED': 11001,
    'CONNECTION_INVALID': 11002,
    'SUBSCRIPTION_FAILED': 11011,
    'UNSUBSCRIPTION_FAILED': 11012,
    'SNAPSHOT_FAILED': 11013,
    'CHANNELP_FAILED': 11031,
    'CHANNELR_FAILED': 11032
}


def DataType(c, d):
    return {"name": c, "type": d}


def enable_log(a):
    global HSD_Flag
    HSD_Flag = a


TopicTypes = {
    "SCRIP": "sf",
    "INDEX": "if",
    "DEPTH": "dp"
}
INDEX_INDEX = {
    "LTP": 2,
    "CLOSE": 3,
    "CHANGE": 10,
    "PERCHANGE": 11,
    "MULTIPLIER": 8,
    "PRECISION": 9
}
SCRIP_INDEX = {
    "VOLUME": 4,
    "LTP": 5,
    "CLOSE": 21,
    "VWAP": 13,
    "MULTIPLIER": 23,
    "PRECISION": 24,
    "CHANGE": 25,
    "PERCHANGE": 26,
    "TURNOVER": 27
}
Keys = {
    "TYPE": "type",
    "USER_ID": "user",
    "SESSION_ID": "sessionid",
    "SCRIPS": "scrips",
    "CHANNEL_NUM": "channelnum",
    "CHANNEL_NUMS": "channelnums",
    "JWT": "jwt",
    "REDIS_KEY": "redis",
    "STK_PRC": "stkprc",
    "HIGH_STK": "highstk",
    "LOW_STK": "lowstk",
    "OPC_KEY": "key",
    "AUTHORIZATION": "Authorization",
    "SID": "Sid",
    "X_ACCESS_TOKEN": "x-access-token",
    "SOURCE": "source"
}
ReqTypeValues = {
    "CONNECTION": "cn",
    "SCRIP_SUBS": "mws",
    "SCRIP_UNSUBS": "mwu",
    "INDEX_SUBS": "ifs",
    "INDEX_UNSUBS": "ifu",
    "DEPTH_SUBS": "dps",
    "DEPTH_UNSUBS": "dpu",
    "CHANNEL_RESUME": "cr",
    "CHANNEL_PAUSE": "cp",
    "SNAP_MW": "mwsp",
    "SNAP_DP": "dpsp",
    "SNAP_IF": "ifsp",
    "OPC_SUBS": "opc",
    "THROTTLING_INTERVAL": "ti",
    "STR": "str",
    "FORCE_CONNECTION": "fcn",
    "LOG": "log"
}

INDEX_MAPPING = [None] * 55
INDEX_MAPPING[0] = DataType("ftm0", FieldTypes.get("DATE"))
INDEX_MAPPING[1] = DataType("dtm1", FieldTypes.get("DATE"))
INDEX_MAPPING[INDEX_INDEX["LTP"]] = DataType("iv", FieldTypes.get("FLOAT32"))
INDEX_MAPPING[INDEX_INDEX["CLOSE"]] = DataType("ic", FieldTypes.get("FLOAT32"))
INDEX_MAPPING[4] = DataType("tvalue", FieldTypes.get("DATE"))
INDEX_MAPPING[5] = DataType("highPrice", FieldTypes.get("FLOAT32"))
INDEX_MAPPING[6] = DataType("lowPrice", FieldTypes.get("FLOAT32"))
INDEX_MAPPING[7] = DataType("openingPrice", FieldTypes.get("FLOAT32"))
INDEX_MAPPING.append(DataType("mul", FieldTypes.get("LONG")))
INDEX_MAPPING[INDEX_INDEX["PRECISION"]] = DataType("prec", FieldTypes.get("LONG"))
INDEX_MAPPING[INDEX_INDEX["CHANGE"]] = DataType("cng", FieldTypes.get("FLOAT32"))
INDEX_MAPPING[INDEX_INDEX["PERCHANGE"]] = DataType("nc", FieldTypes.get("STRING"))
INDEX_MAPPING[STRING_INDEX["NAME"]] = DataType("name", FieldTypes.get("STRING"))
INDEX_MAPPING[STRING_INDEX["SYMBOL"]] = DataType("tk", FieldTypes.get("STRING"))
INDEX_MAPPING[STRING_INDEX["EXCHG"]] = DataType("e", FieldTypes.get("STRING"))
INDEX_MAPPING[STRING_INDEX["TSYMBOL"]] = DataType("ts", FieldTypes.get("STRING"))

SCRIP_MAPPING = [None] * 100
SCRIP_MAPPING[0] = DataType("ftm0", FieldTypes["DATE"])
SCRIP_MAPPING[1] = DataType("dtm1", FieldTypes["DATE"])
SCRIP_MAPPING[2] = DataType("fdtm", FieldTypes["DATE"])
SCRIP_MAPPING[3] = DataType("ltt", FieldTypes["DATE"])
SCRIP_MAPPING[SCRIP_INDEX["VOLUME"]] = DataType("v", FieldTypes["LONG"])
SCRIP_MAPPING[SCRIP_INDEX["LTP"]] = DataType("ltp", FieldTypes["FLOAT32"])
SCRIP_MAPPING[6] = DataType("ltq", FieldTypes["LONG"])
SCRIP_MAPPING[7] = DataType("tbq", FieldTypes["LONG"])
SCRIP_MAPPING[8] = DataType("tsq", FieldTypes["LONG"])
SCRIP_MAPPING[9] = DataType("bp", FieldTypes["FLOAT32"])
SCRIP_MAPPING[10] = DataType("sp", FieldTypes["FLOAT32"])
SCRIP_MAPPING[11] = DataType("bq", FieldTypes["LONG"])
SCRIP_MAPPING[12] = DataType("bs", FieldTypes["LONG"])
SCRIP_MAPPING[SCRIP_INDEX["VWAP"]] = DataType("ap", FieldTypes["FLOAT32"])
SCRIP_MAPPING[14] = DataType("lo", FieldTypes["FLOAT32"])
SCRIP_MAPPING[15] = DataType("h", FieldTypes["FLOAT32"])
SCRIP_MAPPING[16] = DataType("lcl", FieldTypes["FLOAT32"])
SCRIP_MAPPING[17] = DataType("ucl", FieldTypes["FLOAT32"])
SCRIP_MAPPING[18] = DataType("yh", FieldTypes["FLOAT32"])
SCRIP_MAPPING[19] = DataType("yl", FieldTypes["FLOAT32"])
SCRIP_MAPPING[20] = DataType("op", FieldTypes["FLOAT32"])
SCRIP_MAPPING[SCRIP_INDEX["CLOSE"]] = DataType("c", FieldTypes["FLOAT32"])
SCRIP_MAPPING[22] = DataType("oi", FieldTypes["LONG"])
SCRIP_MAPPING[SCRIP_INDEX["MULTIPLIER"]] = DataType("mul", FieldTypes["LONG"])
SCRIP_MAPPING[SCRIP_INDEX["PRECISION"]] = DataType("prec", FieldTypes["LONG"])
SCRIP_MAPPING[SCRIP_INDEX["CHANGE"]] = DataType("cng", FieldTypes["FLOAT32"])
SCRIP_MAPPING[SCRIP_INDEX["PERCHANGE"]] = DataType("nc", FieldTypes["STRING"])
SCRIP_MAPPING[SCRIP_INDEX["TURNOVER"]] = DataType("to", FieldTypes["FLOAT32"])
SCRIP_MAPPING[STRING_INDEX["NAME"]] = DataType("name", FieldTypes["STRING"])
SCRIP_MAPPING[STRING_INDEX["SYMBOL"]] = DataType("tk", FieldTypes["STRING"])
SCRIP_MAPPING[STRING_INDEX["EXCHG"]] = DataType("e", FieldTypes["STRING"])
SCRIP_MAPPING[STRING_INDEX["TSYMBOL"]] = DataType("ts", FieldTypes["STRING"])

DEPTH_MAPPING = [None] * 55
DEPTH_MAPPING[0] = (DataType("ftm0", FieldTypes.get("DATE")))
DEPTH_MAPPING[1] = (DataType("dtm1", FieldTypes.get("DATE")))
DEPTH_MAPPING[2] = (DataType("bp", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[3] = (DataType("bp1", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[4] = (DataType("bp2", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[5] = (DataType("bp3", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[6] = (DataType("bp4", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[7] = (DataType("sp", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[8] = (DataType("sp1", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[9] = (DataType("sp2", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[10] = (DataType("sp3", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[11] = (DataType("sp4", FieldTypes.get("FLOAT32")))
DEPTH_MAPPING[12] = (DataType("bq", FieldTypes.get("LONG")))
DEPTH_MAPPING[13] = (DataType("bq1", FieldTypes.get("LONG")))
DEPTH_MAPPING[14] = (DataType("bq2", FieldTypes.get("LONG")))
DEPTH_MAPPING[15] = (DataType("bq3", FieldTypes.get("LONG")))
DEPTH_MAPPING[16] = (DataType("bq4", FieldTypes.get("LONG")))
DEPTH_MAPPING[17] = (DataType("bs", FieldTypes.get("LONG")))
DEPTH_MAPPING[18] = (DataType("bs1", FieldTypes.get("LONG")))
DEPTH_MAPPING[19] = (DataType("bs2", FieldTypes.get("LONG")))
DEPTH_MAPPING[20] = (DataType("bs3", FieldTypes.get("LONG")))
DEPTH_MAPPING[21] = (DataType("bs4", FieldTypes.get("LONG")))
DEPTH_MAPPING[22] = (DataType("bno1", FieldTypes.get("LONG")))
DEPTH_MAPPING[23] = (DataType("bno2", FieldTypes.get("LONG")))
DEPTH_MAPPING[24] = (DataType("bno3", FieldTypes.get("LONG")))
DEPTH_MAPPING[25] = (DataType("bno4", FieldTypes.get("LONG")))
DEPTH_MAPPING[26] = (DataType("bno5", FieldTypes.get("LONG")))
DEPTH_MAPPING[27] = (DataType("sno1", FieldTypes.get("LONG")))
DEPTH_MAPPING[28] = (DataType("sno2", FieldTypes.get("LONG")))
DEPTH_MAPPING[29] = (DataType("sno3", FieldTypes.get("LONG")))
DEPTH_MAPPING[30] = (DataType("sno4", FieldTypes.get("LONG")))
DEPTH_MAPPING[31] = (DataType("sno5", FieldTypes.get("LONG")))
DEPTH_MAPPING[DEPTH_INDEX["MULTIPLIER"]] = DataType("mul", FieldTypes["LONG"])
DEPTH_MAPPING[DEPTH_INDEX["PRECISION"]] = DataType("prec", FieldTypes["LONG"])
DEPTH_MAPPING[STRING_INDEX["NAME"]] = DataType("name", FieldTypes["STRING"])
DEPTH_MAPPING[STRING_INDEX["SYMBOL"]] = DataType("tk", FieldTypes["STRING"])
DEPTH_MAPPING[STRING_INDEX["EXCHG"]] = DataType("e", FieldTypes["STRING"])
DEPTH_MAPPING[STRING_INDEX["TSYMBOL"]] = DataType("ts", FieldTypes["STRING"])


def leadingZero(a):
    return "0" + str(a) if a < 10 else str(a)


def getFormatDate(a):
    date = datetime.datetime.fromtimestamp(a)
    formatDate = "{}/{}/{} {}:{}:{}".format(
        leadingZero(date.day),
        leadingZero(date.month),
        date.year,
        leadingZero(date.hour),
        leadingZero(date.minute),
        leadingZero(date.second)
    )
    return formatDate


class ByteData:
    def __init__(self, c):
        self.pos = 0
        self.bytes = [0] * (c)
        self.startOfMsg = 0

    def lenth(self):
        # print("lenght of the bytes", self.bytes, len(self.bytes))
        pass

    def markStartOfMsg(self):
        self.startOfMsg = self.pos
        self.pos += 2

    def markEndOfMsg(self):
        len = (self.pos - self.startOfMsg - 2)
        self.bytes[0] = ((len >> 8) & 255)
        self.bytes[1] = (len & 255)

    def clear(self):
        self.pos = 0

    def getPosition(self):
        return self.pos

    def getBytes(self):
        return self.bytes

    def appendByte(self, d):
        # print("in append Bytes POS", self.pos)
        # print("in append Bytes d", d)
        self.bytes[self.pos] = d
        self.pos += 1
        # print("in append Bytes", self.bytes)

    def appendByteAtPos(self, e, d):
        self.bytes[e] = d

    def appendChar(self, d):
        self.bytes[self.pos] = d
        self.pos += 1

    def appendCharAtPos(self, e, d):
        self.bytes[e] = d

    def appendShort(self, d):
        self.bytes[self.pos] = ((d >> 8) & 255)
        self.pos += 1
        self.bytes[self.pos] = (d & 255)
        self.pos += 1

    def appendInt(self, d):
        self.bytes[self.pos] = ((d >> 24) & 255)
        self.pos += 1
        self.bytes[self.pos] = ((d >> 16) & 255)
        self.pos += 1
        self.bytes[self.pos] = ((d >> 8) & 255)
        self.pos += 1
        self.bytes[self.pos] = (d & 255)
        self.pos += 1

    def appendLong(self, d):
        self.bytes[self.pos] = ((d >> 56) & 255)
        self.pos += 1
        self.bytes[self.pos] = ((d >> 48) & 255)
        self.pos += 1
        self.bytes[self.pos] = ((d >> 40) & 255)
        self.pos += 1
        self.bytes[self.pos] = ((d >> 32) & 255)
        self.pos += 1
        self.bytes[self.pos] = ((d >> 24) & 255)
        self.pos += 1
        self.bytes[self.pos] = ((d >> 16) & 255)
        self.pos += 1
        self.bytes[self.pos] = ((d >> 8) & 255)
        self.pos += 1
        self.bytes[self.pos] = (d & 255)
        self.pos += 1

    def append_long_as_big_int(self, e):
        d = int(e)
        self.bytes.append((d >> 56) & 255)
        self.bytes.append((d >> 48) & 255)
        self.bytes.append((d >> 40) & 255)
        self.bytes.append((d >> 32) & 255)
        self.bytes.append((d >> 24) & 255)
        self.bytes.append((d >> 16) & 255)
        self.bytes.append((d >> 8) & 255)
        self.bytes.append(d & 255)

    def append_string(self, d):
        str_len = len(d)
        for i in range(str_len):
            self.bytes[self.pos] = ord(d[i])
            self.pos += 1
            # self.bytes.append(ord(d[i]))

    def append_byte_array(self, d):
        byte_len = len(d)
        for i in range(byte_len):
            self.bytes[self.pos] = d[i]
            self.pos += 1
            # self.bytes.append(d[i])

    def appendByteArr(self, e, d):
        for i in range(d):
            self.bytes[self.pos] = e[i]
            self.pos += 1


class TopicData:
    def __init__(self, feed_type):
        self.feedType = feed_type
        self.exchange = None
        self.symbol = None
        self.tSymbol = None
        self.multiplier = 1
        self.precision = 2
        self.precisionValue = 100
        self.jsonArray = None
        self.fieldDataArray = [None] * 100
        self.updatedFieldsArray = [None] * 100
        self.fieldDataArray[STRING_INDEX["NAME"]] = feed_type

    def getKey(self):
        return f"{self.exchange}|{self.symbol}"

    def setLongValues(self, index_val, value):
        if self.fieldDataArray[index_val] != value and value != TRASH_VAL:
            self.fieldDataArray[index_val] = value
            self.updatedFieldsArray[index_val] = True

    def prepareCommonData(self):
        self.updatedFieldsArray[STRING_INDEX["NAME"]] = True
        self.updatedFieldsArray[STRING_INDEX["EXCHG"]] = True
        self.updatedFieldsArray[STRING_INDEX["SYMBOL"]] = True

    def setStringValues(self, e, d):
        if e == STRING_INDEX["SYMBOL"]:
            self.symbol = d
            self.fieldDataArray[STRING_INDEX["SYMBOL"]] = d
        elif e == STRING_INDEX["EXCHG"]:
            self.exchange = d
            self.fieldDataArray[STRING_INDEX["EXCHG"]] = d
        elif e == STRING_INDEX["TSYMBOL"]:
            self.tSymbol = d
            self.fieldDataArray[STRING_INDEX["TSYMBOL"]] = d
            self.updatedFieldsArray[STRING_INDEX["TSYMBOL"]] = True


class DepthTopicData(TopicData):
    def __init__(self):
        # print("INSIDE DepthTopicData")
        super().__init__(TopicTypes["DEPTH"])
        self.updatedFieldsArray = [None] * 100
        self.multiplier = None
        self.precision = None
        self.precisionValue = None

    def setMultiplierAndPrec(self):
        # print("INTO setMultiplierAndPrec")
        if self.updatedFieldsArray[DEPTH_INDEX['PRECISION']]:
            self.precision = self.fieldDataArray[DEPTH_INDEX['PRECISION']]
            self.precisionValue = 10 ** self.precision
        if self.updatedFieldsArray[DEPTH_INDEX['MULTIPLIER']]:
            self.multiplier = self.fieldDataArray[DEPTH_INDEX['MULTIPLIER']]

    def prepareData(self, type=None):
        # print("INSIDE prepareData")
        self.prepareCommonData()
        # print("\nDepth:", self.feedType, self.exchange, self.symbol)
        json_res = {}
        for d in range(len(DEPTH_MAPPING)):
            c = DEPTH_MAPPING[d]
            e = self.fieldDataArray[d]
            if self.updatedFieldsArray[d] and e is not None and c:
                if c["type"] == FieldTypes.get("FLOAT32"):
                    e = round(e / (self.multiplier * self.precisionValue), self.precision)
                elif c["type"] == FieldTypes.get("DATE"):
                    e = getFormatDate(e)
                # print(d, ":", c["name"], ":", e)
                json_res[c["name"]] = str(e)
        self.updatedFieldsArray = [None] * 100
        # print("INSIDE Parse Data", json_res)
        if type is not None:
            json_res["request_type"] = type

        return json_res


def get_acknowledgement_req(a):
    buffer = ByteData(11) #bytearray(11)
    buffer.markStartOfMsg()
    buffer.appendByte(BinRespTypes["ACK_TYPE"])
    buffer.appendByte(1)
    buffer.appendByte(1)
    buffer.appendShort(4)
    buffer.appendInt(a)
    buffer.markEndOfMsg()
    return buffer.getBytes()


def prepare_connection_request(a):
    user_id_len = len(a)
    src = "JS_API"
    src_len = len(src)
    buffer = bytearray(user_id_len + src_len + 10)
    buffer[0] = BinRespTypes.get("CONNECTION_TYPE")
    buffer[1] = 2
    buffer[2] = 1
    buffer[3:5] = int(user_id_len).to_bytes(2, byteorder='big')
    buffer[5:5 + user_id_len] = a.encode()
    buffer[5 + user_id_len] = 2
    buffer[6 + user_id_len:8 + user_id_len] = int(src_len).to_bytes(2, byteorder='big')
    buffer[8 + user_id_len:8 + user_id_len + src_len] = src.encode()
    buffer[8 + user_id_len + src_len] = BinRespTypes.get("END_OF_MSG")
    return buffer


def prepareConnectionRequest2(a, c):
    # a = bytearray(bytes(a, encoding='utf8'))
    # c = bytearray(bytes(c, encoding='utf8'))
    src = "JS_API"
    # src = bytearray(bytes(src, encoding='utf8'))
    srcLen = len(src)
    jwtLen = len(a)
    redisLen = len(c)
    buffer = ByteData(srcLen + jwtLen + redisLen + 13)
    buffer.markStartOfMsg()
    buffer.appendByte(BinRespTypes["CONNECTION_TYPE"])
    buffer.appendByte(3)
    buffer.appendByte(1)
    buffer.appendShort(jwtLen)
    buffer.append_string(a)
    buffer.appendByte(2)
    buffer.appendShort(redisLen)
    buffer.append_string(c)
    buffer.appendByte(3)
    buffer.appendShort(srcLen)
    buffer.append_string(src)
    buffer.markEndOfMsg()
    return buffer.getBytes()


def is_scrip_ok(a):
    scrips_count = len(a.split("&"))
    if scrips_count > MAX_SCRIPS:
        print("Maximum scrips allowed per request is " + str(MAX_SCRIPS))
        return False
    return True


def getScripByteArray(c, a):
    if c[-1] == "&":
        c = c[:-1]
    scripArray = c.split("&")
    scripsCount = len(scripArray)
    dataLen = 0
    for index in range(scripsCount):
        scripArray[index] = a + "|" + scripArray[index]
        dataLen += len(scripArray[index]) + 1
    # print("Data len ", dataLen)
    bytes = [0] * (dataLen + 2)
    pos = 0
    bytes[pos] = (scripsCount >> 8) & 255
    pos += 1
    bytes[pos] = scripsCount & 255
    pos += 1
    for index in range(scripsCount):
        currScrip = scripArray[index]
        scripLen = len(currScrip)
        bytes[pos] = scripLen & 255
        pos += 1
        for strIndex in range(scripLen):
            bytes[pos] = ord(currScrip[strIndex])
            pos += 1
    # print("Bytes ", bytes)
    return bytes


def prepareSubsUnSubsRequest(scrips, subscribe_type, scrip_prefix, channel_num):
    # print("Prepare prepareSubsUnSubsRequest")
    if not is_scrip_ok(scrips):
        return

    dataArr = getScripByteArray(scrips, scrip_prefix)
    # print("Length Arr", dataArr)
    # buffer = [0] * (len(dataArr) + 11) #ByteData(len(dataArr) + 11)
    buffer = ByteData(len(dataArr) + 11)
    buffer.markStartOfMsg()
    buffer.appendByte(subscribe_type)
    buffer.appendByte(2)
    buffer.appendByte(1)
    buffer.appendShort(len(dataArr))
    buffer.appendByteArr(dataArr, len(dataArr))
    buffer.appendByte(2)
    buffer.appendShort(1)
    buffer.appendByte(int(channel_num))
    buffer.markEndOfMsg()
    return buffer.getBytes()


def prepareSnapshotRequest(a, c, d):
    # print("INTO prepareSnapshotRequest", a, c, d)
    if not is_scrip_ok(a):
        return
    dataArr = getScripByteArray(a, d)
    # print("DATA ARRAY", dataArr)
    buffer = ByteData(len(dataArr) + 7)
    buffer.markStartOfMsg()
    buffer.appendByte(c)
    buffer.appendByte(1)
    buffer.appendByte(2)
    buffer.appendShort(len(dataArr))
    buffer.appendByteArr(dataArr, len(dataArr))
    buffer.markEndOfMsg()
    return buffer.getBytes()


def prepareChannelRequest(c, a):
    buffer = bytearray(15)
    buffer[0] = c
    buffer[1] = 1
    buffer[2] = 1
    buffer[3:5] = (8).to_bytes(2, byteorder='big')
    int1, int2 = 0, 0
    for d in a:
        if 0 < d <= 32:
            int1 |= 1 << d
        elif 32 < d <= 64:
            int2 |= 1 << d
        else:
            print("Error: Channel values must be in this range  [ val > 0 && val < 65 ]")
    buffer[5:9] = int2.to_bytes(4, byteorder='big')
    buffer[9:13] = int1.to_bytes(4, byteorder='big')
    return buffer


def prepareThrottlingIntervalRequest(a):
    buffer = bytearray(11)
    buffer[0] = BinRespTypes.get("THROTTLING_TYPE")
    buffer[1] = 1
    buffer[2] = 1
    buffer[3] = (4 >> 8) & 255
    buffer[4] = 4 & 255
    buffer[5] = (a >> 24) & 255
    buffer[6] = (a >> 16) & 255
    buffer[7] = (a >> 8) & 255
    buffer[8] = a & 255
    return buffer


def get_scrip_byte_array(c, a):
    if c[-1] == "&":
        c = c[:-1]
    scrip_array = c.split("&")
    scrips_count = len(scrip_array)
    data_len = 0
    for index in range(scrips_count):
        scrip_array[index] = a + "|" + scrip_array[index]
        data_len += len(scrip_array[index]) + 1
    bytes = bytearray(data_len + 2)
    pos = 0
    bytes[pos] = (scrips_count >> 8) & 255
    pos += 1
    bytes[pos] = scrips_count & 255
    pos += 1
    for index in range(scrips_count):
        curr_scrip = scrip_array[index]
        scrip_len = len(curr_scrip)
        bytes[pos] = scrip_len & 255
        pos += 1
        for str_index in range(scrip_len):
            bytes[pos] = ord(curr_scrip[str_index])
            pos += 1
    return bytes


def get_opc_chain_subs_request(d, e, a, c, f):
    opc_key_len = len(d)
    buffer = bytearray(opc_key_len + 30)
    pos = 0
    buffer[pos] = BinRespTypes.get("OPC_SUBSCRIBE")
    pos += 1
    buffer[pos] = 5
    pos += 1
    buffer[pos] = 1
    pos += 1
    buffer[pos] = opc_key_len >> 8 & 255
    pos += 1
    buffer[pos] = opc_key_len & 255
    pos += 1
    for i in range(opc_key_len):
        buffer[pos] = ord(d[i])
        pos += 1
    buffer[pos] = 2
    pos += 1
    buffer[pos] = 8 >> 8 & 255
    pos += 1
    buffer[pos] = 8 & 255
    pos += 1
    # The below code assumes the input value of e is a 64-bit integer
    buffer[pos] = e >> 56 & 255
    pos += 1
    buffer[pos] = e >> 48 & 255
    pos += 1
    buffer[pos] = e >> 40 & 255
    pos += 1
    buffer[pos] = e >> 32 & 255
    pos += 1
    buffer[pos] = e >> 24 & 255
    pos += 1
    buffer[pos] = e >> 16 & 255
    pos += 1
    buffer[pos] = e >> 8 & 255
    pos += 1
    buffer[pos] = e & 255
    pos += 1
    buffer[pos] = 3
    pos += 1
    buffer[pos] = 1 >> 8 & 255
    pos += 1
    buffer[pos] = 1 & 255
    pos += 1
    buffer[pos] = a
    pos += 1
    buffer[pos] = 4
    pos += 1
    buffer[pos] = 1 >> 8 & 255
    pos += 1
    buffer[pos] = 1 & 255
    pos += 1
    buffer[pos] = c
    pos += 1
    buffer[pos] = 5
    pos += 1
    buffer[pos] = 1 >> 8 & 255
    pos += 1
    buffer[pos] = 1 & 255
    pos += 1
    buffer[pos] = f
    return buffer


def send_json_arr_resp(a):
    json_arr_res = []
    json_arr_res.append(a)
    return json.dumps(json_arr_res)


def buf2long(a):
    b = bytearray(a)
    val = 0
    leng = len(b)
    for i in range(leng):
        j = leng - 1 - i
        val += b[j] << (i * 8)
    # return val
    return val if val < 2 ** 31 else val - 2 ** 32


def buf2string(a):
    import numpy as np
    return ''.join(map(chr, np.frombuffer(a, dtype=np.uint8)))


class ScripTopicData(TopicData):
    def __init__(self):
        super().__init__(TopicTypes["SCRIP"])
        # print("After topic")
        self.precision = None
        self.precisionValue = None
        self.multiplier = None

    def setMultiplierAndPrec(self):
        if self.updatedFieldsArray[SCRIP_INDEX["PRECISION"]]:
            self.precision = self.fieldDataArray[SCRIP_INDEX["PRECISION"]]
            self.precisionValue = pow(10, self.precision)
        if self.updatedFieldsArray[SCRIP_INDEX["MULTIPLIER"]]:
            self.multiplier = self.fieldDataArray[SCRIP_INDEX["MULTIPLIER"]]

    def prepareData(self,type=None):
        self.prepareCommonData()
        #hardcoded formatting is removed and made it dynamic
        precesionFormat="{:."+str(self.precision)+"f}"
        if self.updatedFieldsArray[SCRIP_INDEX["LTP"]] or self.updatedFieldsArray[SCRIP_INDEX["CLOSE"]]:
            ltp = self.fieldDataArray[SCRIP_INDEX["LTP"]]
            close = self.fieldDataArray[SCRIP_INDEX["CLOSE"]]
            if ltp is not None and close is not None:
                change = ltp - close
                self.fieldDataArray[SCRIP_INDEX["CHANGE"]] = change
                self.updatedFieldsArray[SCRIP_INDEX["CHANGE"]] = True
                self.fieldDataArray[SCRIP_INDEX["PERCHANGE"]] = precesionFormat.format((change / close * 100))
                self.updatedFieldsArray[SCRIP_INDEX["PERCHANGE"]] = True
        if self.updatedFieldsArray[SCRIP_INDEX["VOLUME"]] or self.updatedFieldsArray[SCRIP_INDEX["VWAP"]]:
            volume = self.fieldDataArray[SCRIP_INDEX["VOLUME"]]
            vwap = self.fieldDataArray[SCRIP_INDEX["VWAP"]]
            if volume is not None and vwap is not None:
                self.fieldDataArray[SCRIP_INDEX["TURNOVER"]] = volume * vwap
                self.updatedFieldsArray[SCRIP_INDEX["TURNOVER"]] = True
        # print("\nScrip::" + self.feedType + "|" + self.exchange + "|" + self.symbol)
        jsonRes = {}
        for index in range(len(SCRIP_MAPPING)):
            dataType = SCRIP_MAPPING[index]
            val = self.fieldDataArray[index]
            if self.updatedFieldsArray[index] and val is not None and dataType:
                if dataType["type"] == FieldTypes["FLOAT32"]:

                    val = precesionFormat.format(val / (self.multiplier * self.precisionValue))
                    # val = "{:.4f}".format(val / (self.multiplier * self.precisionValue))
                elif dataType["type"] == FieldTypes["DATE"]:
                    val = getFormatDate(val)
                # print(str(index) + ":" + dataType["name"] + ":" + str(val))
                jsonRes[dataType["name"]] = str(val)
        self.updatedFieldsArray = [None] * 100
        if type is not None:
            jsonRes["request_type"]=type

        return jsonRes


class IndexTopicData(TopicData):
    def __init__(self):
        # print("INSIDE IndexTopicData")
        super().__init__(TopicTypes["INDEX"])
        self.updatedFieldsArray = [None] * 100
        self.multiplier = None
        self.precision = None
        self.precisionValue = None

    def setMultiplierAndPrec(self):
        if self.updatedFieldsArray[INDEX_INDEX["PRECISION"]]:
            self.precision = self.fieldDataArray[INDEX_INDEX["PRECISION"]]
            self.precisionValue = 10 ** self.precision
        if self.updatedFieldsArray[INDEX_INDEX["MULTIPLIER"]]:
            self.multiplier = self.fieldDataArray[INDEX_INDEX["MULTIPLIER"]]

    def prepareData(self, type=None):
        self.prepareCommonData()
        if self.updatedFieldsArray[INDEX_INDEX["LTP"]] or self.updatedFieldsArray[INDEX_INDEX["CLOSE"]]:
            ltp = self.fieldDataArray[INDEX_INDEX["LTP"]]
            close = self.fieldDataArray[INDEX_INDEX["CLOSE"]]
            if ltp is not None and close is not None:
                change = ltp - close
                self.fieldDataArray[INDEX_INDEX["CHANGE"]] = change
                self.updatedFieldsArray[INDEX_INDEX["CHANGE"]] = True
                per_change = round(change / close * 100, self.precision)
                self.fieldDataArray[INDEX_INDEX["PERCHANGE"]] = per_change
                self.updatedFieldsArray[INDEX_INDEX["PERCHANGE"]] = True
        # print("\nIndex::" + self.feedType + "|" + self.exchange + "|" + self.symbol)
        json_res = {}
        for index in range(len(INDEX_MAPPING)):
            data_type = INDEX_MAPPING[index]
            val = self.fieldDataArray[index]
            if self.updatedFieldsArray[index] and val is not None and data_type is not None:
                if data_type["type"] == FieldTypes["FLOAT32"]:
                    val = round(val / (self.multiplier * self.precisionValue), self.precision)
                elif data_type["type"] == FieldTypes["DATE"]:
                    val = getFormatDate(val)
                # print(str(index) + ":" + data_type["name"] + ":" + str(val))
                json_res[data_type["name"]] = str(val)
        self.updatedFieldsArray = [None] * 100
        if type is not None:
            json_res["request_type"] = type

        return json_res


SCRIP_PREFIX = "sf"
INDEX_PREFIX = "if"
DEPTH_PREFIX = "dp"


def convert_to_dict(scrips=None, channelnum=None):
    dict_data = {"scrips": scrips, "sub_type": BinRespTypes.get("SUBSCRIBE_TYPE"), "SCRIP_PREFIX": SCRIP_PREFIX,
                 "channelnum": channelnum}
    return dict_data

class HSWrapper:
    def __init__(self):
        self.counter = 0
        self.ack_num = 0

    def getNewTopicData(self, c):
        feed_type, *_ = c.split("|")
        topic = None
        if feed_type == TopicTypes.get("SCRIP"):
            topic = ScripTopicData()
        elif feed_type == TopicTypes.get("INDEX"):
            topic = IndexTopicData()
        elif feed_type == TopicTypes.get("DEPTH"):
            topic = DepthTopicData()
        return topic

    def getStatus(self, c, d):
        status = BinRespStat.get("NOT_OK")
        field_count = buf2long(c[d:d + 1])
        d += 1
        if field_count > 0:
            fld = buf2long(c[d:d + 1])
            d = d + 1
            field_length = buf2long(c[d:d + 2])
            d += 2
            status = buf2string(c[d:d + field_length])
            d += field_length
        return status

    def parseData(self, e,ws):
        pos = 0
        # print("INTO Parse Data", e)
        packetsCount = buf2long(e[pos:2])
        pos += 2
        type = int.from_bytes(e[pos:pos + 1], 'big')
        pos += 1
        # print("Type in HSWebsocket", type)
        # print("parse data ", e)
        # print("parse data len", len(e))
        if type == BinRespTypes.get("CONNECTION_TYPE"):
            jsonRes = {}
            fCount = int.from_bytes(e[pos:pos + 1], 'big')
            pos += 1
            if fCount >= 2:
                fid1 = int.from_bytes(e[pos:pos + 1], 'big')
                pos += 1
                valLen = int.from_bytes(e[pos:pos + 2], 'big')
                pos += 2
                status = e[pos:pos + valLen].decode('utf-8')
                pos += valLen
                fid1 = int.from_bytes(e[pos:pos + 1], 'big')
                pos += 1
                valLen = int.from_bytes(e[pos:pos + 2], 'big')
                pos += 2
                ackCount = int.from_bytes(e[pos:pos + valLen], 'big')
                # print("STATUS", status)
                if status == BinRespStat.get("OK"):
                    jsonRes['stat'] = STAT.get("OK")
                    jsonRes['type'] = RespTypeValues.get("CONN")
                    jsonRes['msg'] = "successful"
                    jsonRes['stCode'] = RespCodes.get("SUCCESS")
                elif status == BinRespStat.get("NOT_OK"):
                    jsonRes['stat'] = STAT.get("NOT_OK")
                    jsonRes['type'] = RespTypeValues.get("CONN")
                    jsonRes['msg'] = "failed"
                    jsonRes['stCode'] = RespCodes.get("CONNECTION_FAILED")
                self.ack_num = ackCount
            elif fCount == 1:
                fid1 = int.from_bytes(e[pos:pos + 1], 'big')
                pos += 1
                valLen = int.from_bytes(e[pos:pos + 2], 'big')
                pos += 2
                status = e[pos:pos + valLen].decode('utf-8')
                pos += valLen
                if status == BinRespStat.get("OK"):
                    jsonRes['stat'] = STAT.get("OK")
                    jsonRes['type'] = RespTypeValues.get("CONN")
                    jsonRes['msg'] = "successful"
                    jsonRes['stCode'] = RespCodes.get("SUCCESS")
                elif status == BinRespStat.get("NOT_OK"):
                    jsonRes['stat'] = STAT.get("NOT_OK")
                    jsonRes['type'] = RespTypeValues.get("CONN")
                    jsonRes['msg'] = "failed"
                    jsonRes['stCode'] = RespCodes.get("CONNECTION_FAILED")
            else:
                jsonRes['stat'] = STAT.get("NOT_OK")
                jsonRes['type'] = RespTypeValues.get("CONN")
                jsonRes['msg'] = "invalid field count"
                jsonRes['stCode'] = RespCodes.get("CONNECTION_INVALID")
            return send_json_arr_resp(jsonRes)
        else:
            if type == BinRespTypes.get("DATA_TYPE"):
                # print("IN By Datatype ")
                # print("IN By self.ack_num ", self.ack_num)
                msg_num = 0
                if self.ack_num > 0:
                    # print("ack_num", self.ack_num)
                    self.counter += 1
                    msg_num = buf2long(e[pos: pos + 4])
                    pos += 4
                    if self.counter == self.ack_num:
                        req = get_acknowledgement_req(msg_num)
                        if ws:
                            ws.send(req, 0x2)
                            self.counter = 0
                        # print("Acknowledgement sent for message num:", msg_num)
                h = []
                g = buf2long(e[pos: pos + 2])
                # print("G in ", g)
                pos += 2
                for n in range(g):
                    pos += 2
                    c = buf2long(e[pos: pos + 1])
                    # print("ResponseType:", c)
                    pos += 1
                    if c == ResponseTypes.get("SNAP"):
                        f = buf2long(e[pos: pos + 4])
                        pos += 4
                        # print("topic Id:", f)
                        name_len = buf2long(e[pos: pos + 1])
                        pos += 1
                        topic_name = buf2string(e[pos: pos + name_len])
                        # print("TOPIC Name ", topic_name)
                        pos += name_len
                        d = self.getNewTopicData(topic_name)
                        if d:
                            topic_list[f] = d
                            fcount = buf2long(e[pos: pos + 1])
                            pos += 1
                            for index in range(fcount):
                                fvalue = buf2long(e[pos: pos + 4])
                                d.setLongValues(index, fvalue)
                                pos += 4
                            # print("Able to set ")
                            d.setMultiplierAndPrec()
                            fcount = buf2long(e[pos: pos + 1])
                            pos += 1
                            # print("fcount2:", fcount)
                            for index in range(fcount):
                                fid = buf2long(e[pos: pos + 1])
                                pos += 1
                                data_len = buf2long(e[pos: pos + 1])
                                pos += 1
                                str_val = buf2string(e[pos: pos + data_len])
                                pos += data_len
                                d.setStringValues(fid, str_val)
                                # print(fid, ":", str_val)
                                # print("index:", index, "fid:",fid, "val:", str_val)
                            h.append(d.prepareData("SNAP"))
                        else:
                            print("Invalid topic feed type !")
                    else:
                        if c == ResponseTypes.get("UPDATE"):
                            # print("updates ......")
                            f = buf2long(e[pos: pos + 4])
                            # print("topic Id:", f)
                            pos += 4
                            d = topic_list[f]
                            if not d:
                                print("Topic Not Available in TopicList!")
                            else:
                                # print("INSIDE Else COndition ")
                                fcount = buf2long(e[pos:pos + 1])
                                pos += 1
                                # print("fcount1:", fcount)
                                for index in range(fcount):
                                    fvalue = buf2long(e[pos:pos + 4])
                                    d.setLongValues(index, fvalue)
                                    # d[index] = fvalue
                                    # print("index:", index, "val:", fvalue)
                                    pos += 4
                            h.append(d.prepareData("SUB"))
                        else:
                            print("Invalid ResponseType: " + c)
                # print("Final resoonse ",h)
                return h
            else:
                if type == BinRespTypes.get("SUBSCRIBE_TYPE") or type == BinRespTypes.get("UNSUBSCRIBE_TYPE"):
                    # print("INTO SUBScirbe Condition")
                    status = self.getStatus(e, pos)
                    json_res = {}
                    if status == BinRespStat.get("OK"):
                        json_res["stat"] = STAT.get("OK")
                        json_res[
                            "type"] = RespTypeValues.get("SUBS") if type == BinRespTypes.get(
                            "SUBSCRIBE_TYPE") else RespTypeValues.get("UNSUBS")
                        json_res["msg"] = "successful"
                        json_res["stCode"] = RespCodes.get("SUCCESS")
                    elif status == BinRespStat.get("NOT_OK"):
                        json_res["stat"] = STAT.get("NOT_OK")
                        if type == BinRespTypes.get("SUBSCRIBE_TYPE"):
                            json_res["type"] = RespTypeValues.get("SUBS")
                            json_res["msg"] = "subscription failed"
                            json_res["stCode"] = RespCodes.get("SUBSCRIPTION_FAILED")
                        else:
                            json_res["type"] = RespTypeValues.get("UNSUBS")
                            json_res["msg"] = "unsubscription failed"
                            json_res["stCode"] = RespCodes.get("UNSUBSCRIPTION_FAILED")
                    return send_json_arr_resp(json_res)

                else:
                    if type == BinRespTypes.get("SNAPSHOT"):
                        # print("type is snapshot ")
                        status = self.getStatus(e, pos)
                        json_res = {}
                        if status == BinRespStat.get("OK"):
                            json_res["stat"] = STAT.get("OK")
                            json_res["type"] = RespTypeValues.get("SNAP")
                            json_res["msg"] = "successful"
                            json_res["stCode"] = RespCodes.get("SUCCESS")
                        elif status == BinRespStat.get("NOT_OK"):
                            json_res["stat"] = STAT.get("NOT_OK")
                            json_res["type"] = RespTypeValues.get("SNAP")
                            json_res["msg"] = "failed"
                            json_res["stCode"] = RespCodes.get("SNAPSHOT_FAILED")
                        return send_json_arr_resp(json_res)
                    elif type == BinRespTypes.get("CHPAUSE_TYPE") or type == BinRespTypes.get("CHRESUME_TYPE"):
                        status = self.getStatus(e, pos)
                        json_res = {}
                        if status == BinRespStat.get("OK"):
                            json_res["stat"] = STAT.get("OK")
                            if type == BinRespTypes.get("CHPAUSE_TYPE"):
                                json_res["type"] = RespTypeValues.get("CHANNELP")
                            else:
                                json_res["type"] = RespTypeValues.get("CHANNELR")
                            json_res["msg"] = "successful"
                            json_res["stCode"] = RespCodes.get("SUCCESS")
                        elif status == BinRespStat.get("NOT_OK"):
                            json_res["stat"] = STAT.get("NOT_OK")
                            if type == BinRespTypes.get("CHPAUSE_TYPE"):
                                json_res["type"] = RespTypeValues.get("CHANNELP")
                            else:
                                json_res["type"] = RespTypeValues.get("CHANNELR")
                            json_res["msg"] = "failed"
                            if type == BinRespTypes.get("CHPAUSE_TYPE"):
                                json_res["stCode"] = RespCodes.get("CHANNELP_FAILED")
                            else:
                                json_res["stCode"] = RespCodes.get("CHANNELR_FAILED")
                        return send_json_arr_resp(json_res)
                    elif type == BinRespTypes.get("OPC_SUBSCRIBE"):
                        status = self.getStatus(e, pos)
                        pos += 5
                        json_res = {}
                        if status == BinRespStat.get("OK"):
                            json_res["stat"] = STAT.get("OK")
                            json_res["type"] = RespTypeValues.get("OPC")
                            json_res["msg"] = "successful"
                            json_res["stCode"] = RespCodes.get("SUCCESS")
                            fld = buf2long(e[pos: pos + 1])
                            pos += 1
                            field_length = buf2long(e[pos: pos + 2])
                            pos += 2
                            opc_key = buf2string(e[pos: pos + field_length])
                            pos += field_length
                            json_res["key"] = opc_key
                            fld = buf2long(e[pos: pos + 1])
                            pos += 1
                            field_length = buf2long(e[pos: pos + 2])
                            pos += 2
                            data = buf2string(e[pos:pos + field_length])
                            pos += field_length
                            json_res["scrips"] = json.loads(data)["data"]
                        elif status == BinRespStat.get("NOT_OK"):
                            json_res["stat"] = STAT.get("NOT_OK")
                            json_res["type"] = RespTypeValues.get("OPC")
                            json_res["msg"] = "failed"
                            json_res["stCode"] = 11040

                        return send_json_arr_resp(json_res)
                    else:
                        return None





import os
import duckdb
import pandas as pd 
import datetime
import traceback
import copy
import time
from .dependencies import SparkLib
from .historicalData import HistoricalData

class ADD_INSTRUMENTS():
    def __init__(self) -> None:
        self.intrumentsList = []
    
    def __call__(self, name, exchange, getSpot= True, getOptions=True, optExp=[0,1,2,3,4], getFutures=True, futExp=[0,1], *args, **kwargs):
        self.addInstrument(name, exchange, getSpot, getOptions, optExp, getFutures, futExp)

    def addInstrument(self, name, exchange, getSpot= True, getOptions=True, optExp=[0,1,2,3,4], getFutures=True, futExp=[0,1]):
        obj = locals()
        del obj['self']
        self.intrumentsList.append(obj)

class ManageData():
    def __init__(self, duckFunc):
        self.dataStore = []
        self.duckFunc = duckFunc
    
    def addData(self, data, symbol, token, expiry):
        try: 
            [i.update({"datetime":datetime.datetime.strptime(i['datetime'], "%Y-%m-%d %H:%M:%S")}) for i in data]
            expiry = datetime.datetime.strptime(expiry,"%Y-%m-%dT%H:%M:%S") if expiry else datetime.datetime(1900,1,1)
            [i.update({"symbol":symbol, "token" : token, "expiry" :str(expiry)}) for i in data]
            # self.dataStore.extend(data)
            self.duckFunc(data)
        except:
            print(f"data is not available for {symbol}", traceback.format_exc())

    @property
    def DataFrame(self):
        return pd.DataFrame(self.dataStore)

class FetchData():
    fnoExch = {"NSE":"NFO",
               "BSE":"BFO",
               "MCX":"MCX"}
    
    def __init__(self, accountName, spk:SparkLib, hist:HistoricalData, duckFunc):
        self.accountName=accountName
        self.spk = spk
        self.hist = hist
        self.mg = ManageData(duckFunc=duckFunc)
        
    def __fetchData(self, token, startTime, endTime):
        try: 
            accountName=self.accountName
            data = self.hist.getHistorical(accountName=accountName, token=token, startTime=startTime, endTime=endTime)
            if data['status'] and not data['error'] : 
                return data['data']
            else: 
                print(f"ERROR while fetching data for {token}; message : {data}")
        except Exception as e: 
            print(e, traceback.format_exc())

    def fetchData(self, startTime, startTimeFNO,endTime, instruments:ADD_INSTRUMENTS, sleepInterval:int=0.1):
        for ins in instruments.intrumentsList:
            if ins['getSpot'] : 
                spotToken = self.spk.getTokens(ins['name'], ins['exchange'])['data'][0]
                token = ":".join([spotToken['exch_seg'], spotToken['token']])
                data = self.__fetchData(token=token, startTime=startTime,endTime=endTime)
                self.mg.addData(data, symbol=spotToken['name'], token=token, expiry=None)

            if ins['getOptions'] : 
                optExp = self.spk.getExpiry(ins['name'], self.fnoExch[ins['exchange']], instrument="OPT")['data']
                optExp = [optExp[i] for i in ins["optExp"]]
                for exp in optExp: 
                    tokens = self.spk.getTokens(symbol=ins['name'], exchange=self.fnoExch[ins['exchange']],
                                                expiry=exp, instrument="OPT")
                    for tk in tokens['data']: 
                        data = self.__fetchData(token=":".join([tk['exch_seg'], tk['token']]), 
                                              startTime=startTimeFNO,
                                              endTime=endTime)
                        self.mg.addData(data, symbol=tk['symbol'], token=":".join([tk['exch_seg'], tk['token']]), expiry=tk['expiry'])
                        time.sleep(sleepInterval)
            
            if ins['getFutures']:
                futExp = self.spk.getExpiry(ins['name'], self.fnoExch[ins['exchange']], instrument="FUT")['data']
                futExp = [futExp[i] for i in ins["futExp"]]
                for exp in futExp: 
                    tokens = self.spk.getTokens(symbol=ins['name'], exchange=self.fnoExch[ins['exchange']],
                                                expiry=exp, instrument="FUT")
                    for tk in tokens['data']: 
                        data = self.__fetchData(token=":".join([tk['exch_seg'], tk['token']]), 
                                              startTime=startTimeFNO,
                                              endTime=endTime)
                        self.mg.addData(data, symbol=tk['symbol'], token=":".join([tk['exch_seg'], tk['token']]), expiry=tk['expiry'])
        return self.mg.dataStore

class buildHist():
    def __init__(self, filename:str, accessToken:str=None, path:str="./"):
        if not accessToken :
            if os.environ.get("MATICALGOS_AccessToken"):
                accessToken = os.environ["MATICALGOS_AccessToken"] 
            else: 
                raise Exception("Please generate access token.")
        self.spk = SparkLib(access_token=accessToken)
        self.hist = HistoricalData(accessToken)
        self.filename = filename
        self.path = path
        self.__createDuckDB(filename=filename, path=path)

    def __createDuckDB(self, filename, path = "."):
        files = os.listdir(path)
        if filename + ".db" not in files: 
            self.conn = duckdb.connect(r"{path}/{filename}.db".format(path=path,filename=filename))
            self.conn.execute("""
                        CREATE TABLE dbtable (
                            datetime TIMESTAMP,
                            open DOUBLE,
                            high DOUBLE,
                            low DOUBLE,
                            close DOUBLE,
                            volume INTEGER,
                            oi INTEGER,
                            symbol VARCHAR,
                            token VARCHAR,
                            expiry TIMESTAMP,
                         )
                            """)
            
            self.conn.execute("""
                        CREATE TABLE dataConfig (
                            name VARCHAR,
                            exchange VARCHAR,
                            getSpot VARCHAR,
                            getOptions VARCHAR,
                            optExp VARCHAR,
                            getFutures VARCHAR,
                            futExp VARCHAR,  
                            lastUpdate TIMESTAMP,
                            maxDays INTEGER,
                         )
                            """)
            self.conn.commit()
            self.conn.close()

    def _conn(self):
        return duckdb.connect(r"{path}/{filename}.db".format(path=self.path,filename=self.filename))
    
    def dumpInstruments(self, instruments:ADD_INSTRUMENTS, maxDays:int):
        conn = self._conn()
        an = copy.deepcopy(instruments.intrumentsList)
        [i.update({"optExp":",".join(map(str,i['optExp'])), "futExp" : ",".join(map(str,i['futExp'])), 
                   "lastUpdate":datetime.datetime.now().replace(hour=23,minute=59) - datetime.timedelta(days=1), 
                   "maxDays" : maxDays}) for i in an]
        an = pd.DataFrame(an)
        conn.execute("""INSERT INTO dataConfig SELECT * FROM an""")
        conn.close()

    def chunks(self, data, batch):
        for i in range(0, len(data), batch):
            yield data[i:i + batch]

    def pushDuckDB(self, data, batch=10000):
        conn = self._conn()
        # for chunk in self.chunks(data, batch):
            # data = pd.DataFrame(chunk)
            # conn.execute("""INSERT INTO dbtable SELECT * FROM data""")
        data = pd.DataFrame(data)
        conn.execute("""INSERT INTO dbtable SELECT * FROM data""")
        conn.commit()
        conn.close()

    def getConf(self):
        conn = self._conn()
        data = conn.execute("""SELECT * FROM dataConfig""").df()
        conn.close()
        return data
    
    def deleteData(self, maxDays):
        conn = self._conn()
        today = datetime.datetime.now().replace(hour=0,minute=0,second=0,microsecond=0)
        endDt = str(today - datetime.timedelta(days=maxDays))
        conn.execute(f"""DELETE FROM dbtable WHERE datetime < '{endDt}'""")
        conn.commit()
        conn.execute(f"""DELETE FROM dbtable WHERE expiry < '{today}' and expiry > '1900-01-01 00:00:00'""")
        conn.commit()
        conn.close()

    def createDataBase(self, AccountName:str, instruments:ADD_INSTRUMENTS, maxDays:int=70, maxDaysFNO:int=14,sleepInterval:int=0.2):
        dData = FetchData(accountName=AccountName,spk=self.spk,hist=self.hist, duckFunc=self.pushDuckDB)
        self.dumpInstruments(instruments=instruments, maxDays=maxDays)
        startTime = (datetime.datetime.now() - datetime.timedelta(days=maxDays)).replace(hour=0,minute=0)
        startTimeFNO = (datetime.datetime.now() - datetime.timedelta(days=maxDaysFNO)).replace(hour=0,minute=0)
        endTime = datetime.datetime.now().replace(hour=23,minute=59) - datetime.timedelta(days=1)
        data = dData.fetchData(startTime=startTime,startTimeFNO=startTimeFNO,endTime=endTime,instruments=instruments,sleepInterval=sleepInterval)
        self.pushDuckDB(data)

    def fillData(self, AccountName:str, sleepInterval:int=0.3):
        dData = FetchData(accountName=AccountName,spk=self.spk,hist=self.hist, duckFunc=self.pushDuckDB)
        conf = self.getConf()
        maxDays = int(conf['maxDays'].iloc[0])
        lastUpdate = conf['lastUpdate'].iloc[0]
        self.deleteData(maxDays)
        endTime = datetime.datetime.now()
        ins = ADD_INSTRUMENTS()
        strBool = lambda x: x.lower() == "true"
        for i in range(len(conf)):
            data = dict(conf.iloc[i])
            data['futExp'] = list(map(int,data['futExp'].split(",")))
            data['optExp'] = list(map(int,data['optExp'].split(",")))
            del data['lastUpdate']
            data = {key: strBool(value) if value in ["true", "false"] else value for key, value in data.items()}
            ins(**data)
        data = dData.fetchData(startTime=lastUpdate,startTimeFNO=lastUpdate,endTime=endTime,instruments=ins,sleepInterval=sleepInterval)
        self.pushDuckDB(data)









import configparser
import os
from datetime import datetime

import socketio


class XTSMDSocket_io(socketio.Client):
    """A Socket.IO client.
    This class implements a fully compliant Socket.IO web client with support
    for websocket and long-polling transports.
    :param reconnection: 'True'. if the client should automatically attempt to
                         reconnect to the server after an interruption, or
                         'False' to not reconnect. The default is 'True'.
    :param reconnection_attempts: How many reconnection attempts to issue
                                  before giving up, or 0 for infinity attempts.
                                  The default is 0.
    :param reconnection_delay: How long to wait in seconds before the first
                               reconnection attempt. Each successive attempt
                               doubles this delay.
    :param reconnection_delay_max: The maximum delay between reconnection
                                   attempts.
    :param randomization_factor: Randomization amount for each delay between
                                 reconnection attempts. The default is 0.5,
                                 which means that each delay is randomly
                                 adjusted by +/- 50%.
    :param logger: To enable logging set to 'True' or pass a logger object to
                   use. To disable logging set to 'False'. The default is
                   'False'.
    :param binary: 'True' to support binary payloads, 'False' to treat all
                   payloads as text. On Python 2, if this is set to 'True',
                   'unicode' values are treated as text, and 'str' and
                   'bytes' values are treated as binary.  This option has no
                   effect on Python 3, where text and binary payloads are
                   always automatically discovered.
    :param json: An alternative json module to use for encoding and decoding
                 packets. Custom json modules must have 'dumps' and 'loads'
                 functions that are compatible with the standard library
                 versions.
    """

    def __init__(self, token, userID,reconnection=True, reconnection_attempts=0, reconnection_delay=1,
                 reconnection_delay_max=50000, randomization_factor=0.5, logger=False, binary=False, json=None,URL=None,
                 **kwargs):
        self.sid = socketio.Client(logger=True, engineio_logger=True)
        self.eventlistener = self.sid

        self.sid.on('connect', self.on_connect)
        self.sid.on('message', self.on_message)

        """Similarly implement partial json full and binary json full."""
        self.sid.on('1501-json-full', self.on_message1501_json_full)
        self.sid.on('1501-json-partial', self.on_message1501_json_partial)

        self.sid.on('1502-json-full', self.on_message1502_json_full)
        self.sid.on('1502-json-partial', self.on_message1502_json_partial)

        self.sid.on('1505-json-full', self.on_message1505_json_full)
        self.sid.on('1505-json-partial', self.on_message1505_json_partial)

        self.sid.on('1507-json-full', self.on_message1507_json_full)

        self.sid.on('1510-json-full', self.on_message1510_json_full)
        self.sid.on('1510-json-partial', self.on_message1510_json_partial)

        self.sid.on('1512-json-full', self.on_message1512_json_full)
        self.sid.on('1512-json-partial', self.on_message1512_json_partial)

        self.sid.on('1105-json-full', self.on_message1105_json_full)
        self.sid.on('1105-json-partial', self.on_message1105_json_partial)

        self.sid.on('disconnect', self.on_disconnect)

        """Get the root url from config file"""
        currDirMain = os.getcwd()
        configParser = configparser.ConfigParser()
        configFilePath = os.path.join(currDirMain, 'config.ini')
        configParser.read(configFilePath)

        # self.port = configParser.get('root_url', 'root')
        self.userID = userID
        publishFormat = 'JSON'
        self.broadcastMode = "Full"#configParser.get('root_url', 'broadcastMode')
        self.token = token
        
        self.port = URL

        port = f'{self.port}/?token='
        # port = f'{self.port}/?token='

        self.connection_url = port + token + '&userID=' + self.userID + '&publishFormat=' + publishFormat + '&broadcastMode=' + self.broadcastMode

    def connect(self, headers={}, transports='websocket', namespaces=None, socketio_path='/apimarketdata/socket.io',
                verify=False):
        """Connect to a Socket.IO server.
        :param verify: Verify SSL
        :param url: The URL of the Socket.IO server. It can include custom
                    query string parameters if required by the server.
        :param headers: A dictionary with custom headers to send with the
                        connection request.
        :param transports: The list of allowed transports. Valid transports
                           are 'polling' and 'websocket'. If not
                           given, the polling transport is connected first,
                           then an upgrade to websocket is attempted.
        :param namespaces: The list of custom namespaces to connect, in
                           addition to the default namespace. If not given,
                           the namespace list is obtained from the registered
                           event handlers.
        :param socketio_path: The endpoint where the Socket.IO server is
                              installed. The default value is appropriate for
                              most cases.

        self.url = self.connection_url
        self.connection_headers = headers
        self.connection_transports = transports
        self.connection_namespaces = namespaces
        self.socketio_path = socketio_path
        
        Connect to the socket.
        """
        url = self.connection_url
        """Connected to the socket."""
        self.sid.connect(url, headers, transports, namespaces, socketio_path)
        self.sid.wait()
        """Disconnected from the socket."""
        # self.sid.disconnect()

    def on_connect(self):
        """Connect from the socket."""
        print('Market Data Socket connected successfully!')

    def on_message(self, data):
        """On receiving message"""
        print('I received a message!' + data)

    def on_message1502_json_full(self, data):
        """On receiving message code 1502 full"""
        # print('I received a 1502 Market depth message!' + data)

    def on_message1507_json_full(self, data):
         """On receiving message code 1507 full"""
         # print('I received a 1507 MarketStatus message!' + data)
         
    def on_message1512_json_full(self, data):
         """On receiving message code 1512 full"""
         # print('I received a 1512 LTP message!' + data)     
 
    def on_message1505_json_full(self, data):
        """On receiving message code 1505 full"""
        # print('I received a 1505 Candle data message!' + data)

    def on_message1510_json_full(self, data):
        """On receiving message code 1510 full"""
        # print('I received a 1510 Open interest message!' + data)

    def on_message1501_json_full(self, data):
        """On receiving message code 1501 full"""
        # print('I received a 1501 Level1,Touchline message!' + data)
    
    def on_message1105_json_full(self, data):
        pass
        # print('I received a 1105, Instrument Property Change Event message!' + data)
    
    def on_message1502_json_partial(self, data):
        """On receiving message code 1502 partial"""
        # print('I received a 1502 partial message!' + data)
    
    def on_message1512_json_partial(self, data):
        """On receiving message code 1512 partial"""
        # print('I received a 1512 LTP message!' + data)

    def on_message1505_json_partial(self, data):
        """On receiving message code 1505 partial"""
        # print('I received a 1505 Candle data message!' + data)

    def on_message1510_json_partial(self, data):
        """On receiving message code 1510 partial"""
        # print('I received a 1510 Open interest message!' + data)

    def on_message1501_json_partial(self, data):
        """On receiving message code 1501 partial"""
        now = datetime.now()
        today = now.strftime("%H:%M:%S")
        # print(today, 'in main 1501 partial Level1,Touchline message!' + data + ' \n')

    def on_message1105_json_partial(self, data):
        """On receiving message code 1105 partial"""
        now = datetime.now()
        today = now.strftime("%H:%M:%S")
        # print(today, 'in main 1105 partial, Instrument Property Change Event!' + data + ' \n')

        # print('I received a 1105 Instrument Property Change Event!' + data)

    def on_disconnect(self):
        """Disconnected from the socket"""
        print('Market Data Socket disconnected!')

    def on_error(self, data):
        """Error from the socket"""
        print('Market Data Error', data)

    def get_emitter(self):
        """For getting the event listener"""
        return self.eventlistener



import requests
import time
import traceback
import datetime
import json

class IiflHist:
    LTPData_per_sec = 1
    MaxQuoteSymbolLimit = 60

    BASE_URL = "https://ttblaze.iifl.com/apimarketdata/"
    ENDPOINTS = {
        "getData" : "instruments/ohlc",
        "getQuotes" : "instruments/quotes"
    }
    HistoricalMapping = {
        "1S" : [1],   # IIfl has 1 minute data ,but no limit on days
        "1" : [60],
        "2" : [120],
        "3" : [180],
        "5" : [300],
        "10" : [600],
        "15" : [900],
        "30" : [1800],
        "60" : [3600],
        "1D" : [86400]
    }

    def __init__(self,sessionData : dict):
        self.sessionid = sessionData['Sessionid']
        self.reqsession = requests.Session()
        self.header = {
            'Content-Type': 'application/json',
            'authorization': self.sessionid
        }
        self.__supportedFormats = list(self.info()['supported timeformats']["timeValues"].values())

    def requestAPI(self, method, url, body = None, is_header = True, timeout = 15):
        try :
            if not is_header :
                resp = self.reqsession.request(method, url, json = body, timeout = timeout)
            else :
                resp = self.reqsession.request(method, url, json = body, timeout = timeout,
                                               headers = self.header)

            data = resp.json()
            if data.get('type') == 'error' :
                return {"status" : False, "data" : [data], "error" : True, "statuscode" : resp.status_code}

            else :
                return {"status" : True, "data" : [data], "error" : False, "statuscode" : resp.status_code}

        except requests.exceptions.Timeout:
            return {"status" : False, "data" : [], "error" : True,
                    "statuscode" : "", "message" : "Timeout error"}

        except Exception as e :
            return {"status" : False, "data" : [], "error" : True,
                    "statuscode" : resp.status_code, "message" : e}

    def get_quotes(self, symbols: list,sleeptime = 0):
        main_data = {}
        tokenMap = {}
        nonTokens = []

        # Map tokens to exchange:symbol and categorize tokens
        for token in symbols:
            histToken = token.get('histToken')
            if histToken:
                tokenMap[histToken] = token['token']
            else:
                nonTokens.append(token['token'])

        if not tokenMap:
            return {
                "status": False,
                "data": [],
                "error": True,
                "message": "No histToken found for any symbols"
            }

        tokenMapKeys = list(tokenMap.keys())
        for i in range(0, len(tokenMapKeys), self.MaxQuoteSymbolLimit):
            symbols = []
            tokens = tokenMapKeys[i:i + self.MaxQuoteSymbolLimit]
            for token in tokens:
                symbols.append({
                    "exchangeSegment": token.split(", ")[0],
                    "exchangeInstrumentID": token.split(", ")[1]
                })
            params = {'instruments': symbols, 'xtsMessageCode': 1501, 'publishFormat': "JSON"}
            url = f'{self.BASE_URL}{self.ENDPOINTS["getQuotes"]}'

            try:
                response = self.requestAPI("POST", url, body=params)
                if not response.get('status') or response.get('error'):
                    return response
                api_data = response.get('data', [])
                if not api_data:
                    return {
                        "status": False,
                        "data": [],
                        "error": True,
                        "message": "No data received"
                    }
                api_data = api_data[0]['result']['listQuotes']
                for message in api_data:
                    message =json.loads(message)
                    symbol = f"{message['ExchangeSegment']}, {message['ExchangeInstrumentID']}"
                    main_data[tokenMap[symbol]] = {
                        'timestamp_str': datetime.datetime.fromtimestamp(
                            int(message['LastUpdateTime'])-19800).strftime('%Y-%m-%d %H:%M:%S'),
                        'timestamp': str(int(message['LastUpdateTime'])-19800),
                        'symbol': tokenMap[symbol],
                        'ltp': message['LastTradedPrice'],
                        'prev_day_close': message['Close'],
                        'oi': 0,
                        'prev_day_oi': 0,
                        'turnover': message['TotalTradedQuantity'],
                        'best_bid_price': message['BidInfo']['Price'],
                        'best_ask_price': message['AskInfo']['Price'],
                        'best_bid_qty': message['BidInfo']['Size'],
                        'best_ask_qty': message['AskInfo']['Size'],
                        'ttq': message['TotalTradedQuantity'],
                        'token': tokenMap[symbol]
                    }
            except Exception as e:
                return {
                    "status": False,
                    "data": [],
                    "error": True,
                    "message": f"Error occurred: {str(e)}"
                }
            time.sleep(sleeptime)

        # Fill in empty data for nonTokens
        for token in nonTokens:
            main_data[token] = {}

        return {
            "status": True,
            "data": main_data,
            "error": False,
            "message": "Data Received"
        }

    def getHistorical(self, token, startTime : datetime, endTime : datetime, interval : str): #,sec = False
        if interval not in self.__supportedFormats:
            return {"status" : False, "error" : True, "message" : f"Invalid interval. Please input as per the format: {self.info()}", "data" : []}

        if not token:
            return {"status" : False, "data" : [], "error" : True, "message" : "No histToken found for the symbol"}

        endTime = min(endTime, datetime.datetime.now())
        main_data = []
        start_date_str = startTime.strftime("%b %d %Y %H%M%S")
        end_date_str = endTime.strftime("%b %d %Y %H%M%S")
        exchangeSegment, exchangeInstrumentID = token.split(", ")
        time_remove = 0 if interval == "1S" else self.HistoricalMapping[interval][0]-1
        try:
            url = f'{self.BASE_URL}{self.ENDPOINTS["getData"]}?exchangeSegment={exchangeSegment}&exchangeInstrumentID={exchangeInstrumentID}&startTime={start_date_str}&endTime={end_date_str}&compressionValue={self.HistoricalMapping[interval][0]}'
            response = self.requestAPI("GET", url)
            if not response.get('status') or response.get('error'):
                return response
            candle_data = response.get('data', [])
            if candle_data and not candle_data[0].get('result', {}).get('dataReponse', '') == '':
                main_data= (self.ConvertToSTD(candle_data[0]['result']['dataReponse'].split(','), time_remove))
            else:
                return {"status": False, "data": [], "error": True, "message": "Data not available"}
        except:
            return {"status": False, "data": [], "error": True, "message": f"Error : {traceback.format_exc()}"}
        
        return {"status": True, "data": main_data, "error": False, "message": "Data Received"} if main_data else \
                {"status": False, "data": [], "error": True, "message": "Data not available"}

    def ConvertToSTD(self, data, request_candle):
        if not data:
            return []
        data = [candle.split('|') for candle in data]
        std_format = [
            {
                'datetime': (datetime.datetime.fromtimestamp(int(candle[0])-request_candle) - datetime.timedelta(hours=5, minutes=30)).strftime("%Y-%m-%d %H:%M:%S"),
                'open': float(candle[1]),
                'high': float(candle[2]),
                'low': float(candle[3]),
                'close': float(candle[4]),
                'volume': int(candle[5]),
                'oi': int(candle[6]) if len(candle) == 7 else 0
            }
            for candle in data
        ]

        return std_format

    def info(self):
        return {"supported timeformats" :
            {"timeValues" :
                {
                    "1 seconds" : "1S",
                    "1 minute" : "1",
                    "2 minute" : "2",
                    "3 minute" : "3",
                    "5 minute"  : "5",
                    "10 minute" : "10",
                    "15 minute" : "15",
                    "30 minute" : "30",
                    "60 minute" : "60",
                    "1 Day" : "1D"
                }}}





from maticalgos.sparkLib import SparkLib

spk = SparkLib(apikeys="")

from maticalgos.zmqWsHandler import ZmqDataWs

if __name__ == "__main__":  
    token = spk.generate_token()
    if token['status'] and not token['error'] : 
        accesstoken = token['data'][0]['access_token']
    else: 
        raise(token)
    zm = ZmqDataWs(accesstoken, _priorityTicks = True)
    zm._zmqConnect()



import json
import time
import websocket
from threading import Thread
import ssl
from .kotakDataBase import *

class KotakDataFeed:
    def __init__(self,token,sid, server_id,
                 on_message=None,
                 on_error=None,
                 on_close=None,
                 on_open=None,
                 retry_attempts=5
                 ):
        self.access_token = token
        self.sid = sid
        self.server_id = server_id
        self.onmessage = on_message
        self.onerror = on_error
        self.onclose = on_close
        self.onopen = on_open

        self.ws = None
        self.max_retry_attempts = retry_attempts if retry_attempts > -1 else 5
        self.current_retry_attempt = 0
        self.rec_delay = 5

        self.hsWrapper = HSWrapper()
        self.un_sub_token = False
        self.quotes_arr = []
        self.sub_list = []
        self.un_sub_list = []
        self.un_sub_channel_token = {}
        self.channel_tokens = {}
        self.live_scrip_type = None
        self.quotes_index = None
        self.un_sub_list_count = 0
        self.un_sub_channel = None
        self.token_limit_reached = False

    def on_message(self, ws, inData):
        outData = None
        if isinstance(inData, bytes):
            jsonData = self.hsWrapper.parseData(inData,ws)
            # print("JSON DATA in HSWEBSOCKE ON MESSAGE", jsonData)
            if jsonData:
                outData = json.dumps(jsonData) if isEncyptOut else jsonData
        else:
            outData = inData if not isEncyptIn else json.loads(inData) if isEncyptOut else inData
        if outData:
            message = outData
            if type(message) == str:
                req_type = json.loads(message)[0]["type"]
                if req_type == 'cn':
                    if len(self.quotes_arr) >= 1:
                        self.call_quotes()
                    if len(self.sub_list) >= 1:
                        self.subscribe_scripts(self.channel_tokens)
                if req_type == "unsub":
                    if len(self.un_sub_channel_token) > 0 and self.un_sub_channel:
                        # remove from sub_list and sub_token
                        self.remove_items(self.un_sub_channel_token[self.un_sub_channel])
                        del self.un_sub_channel_token[self.un_sub_channel]
                    if len(self.un_sub_channel_token) == 0:
                        if self.token_limit_reached:
                            self.sub_list = []
                            self.channel_tokens = {}
                            self.un_sub_channel_token = {}
                    if self.on_message:
                        self.on_message("Un-Subscribed Successfully!")
            elif type(message) == list:

                # print("raw message ",message)

                # print("quotes ",self.quotes_arr)
                request_type=message[0].get('request_type')
                if request_type and request_type == "SNAP" and (len(self.quotes_arr) >= 1):
                    out_list, quote_type = self.quote_response_formatter(message)
                    if len(out_list)>0:
                        # print("length greater than 0 ")
                        quote_message = self.response_format(out_list, quote_type=quote_type)
                        if self.onmessage:
                            self.onmessage({"type": "quotes", "data": quote_message})
                        self.quotes_arr = []
                if len(self.sub_list) >= 1 and self.is_message_for_subscription(message):
                    if self.onmessage:
                        self.onmessage({"type": "stock_feed", "data": message})

                # If there is no other tokens in quotes_arr and sub_list. disconnect the socket
                # print("sublist size ",len(self.sub_list))
                if(len(self.sub_list)<=0):
                    self.close_websocket()

    def on_error(self, ws, error):
        if self.onerror:
            self.onerror(error)
        else:
            print(f'Error : {error}')
        self.retry_connect()

    def on_close(self, ws, close_status_code, close_msg):
        if self.onclose:
            self.onclose(close_status_code, close_msg)
        else:
            print(f'Closed : {close_status_code} : {close_msg}')

    def on_open(self, ws):
        if self.onopen:
            self.onopen()
        else:
            print('Opened')

    def __on_open_callback(self, ws):
        # print("Connection established HSWebSocket")
        req_params = {"type": "cn", "Authorization": self.access_token, "Sid": self.sid}
        self.hs_send(json.dumps(req_params))
        self.on_open(ws)
        self.current_retry_attempt = 0

    def retry_connect(self):
        if self.current_retry_attempt < self.max_retry_attempts:
            self.current_retry_attempt += 1
            time.sleep(self.rec_delay)
            print(f"Retrying {self.current_retry_attempt} of {self.max_retry_attempts}")
            self.start_websocket()
        else:
            print('Max retry attempts reached')

    def ws_run_forever(self):
        self.ws.run_forever(ping_interval=0, sslopt={"cert_reqs": ssl.CERT_NONE}, reconnect= 5)
                            #ping_timeout=1)

    def start_websocket(self):
        try:
            url = f"wss://mlhsm.kotaksecurities.com"
            self.ws = websocket.WebSocketApp(url,
                                             on_message=self.on_message,
                                             on_error=self.on_error,
                                             on_close=self.on_close,
                                             on_open=self.__on_open_callback)
            self.ws_thread = Thread(target=self.ws_run_forever)
            self.ws_thread.daemon = False
            self.ws_thread.start()

        except Exception as e:
            print(f"Error starting websocket: {e}")
            self.retry_connect()

    def close_websocket(self):
        if self.ws:
            self.ws.close()

    def hs_send(self, d):
        req_json = json.loads(d)
        req_type = req_json[Keys.get("TYPE")]
        # print("Req Type", req_type)
        req = {}
        if Keys.get("SCRIPS") in req_json:
            scrips = req_json[Keys.get("SCRIPS")]
            # print("scrips ", scrips)
            channelnum = req_json[Keys.get("CHANNEL_NUM")]
            # print("CHANNEL NUM ", channelnum)
        else:
            scrips = None
            channelnum = 1
        # scrips = None
        # channelnum = req_json[Keys.get("CHANNEL_NUM")]
        if req_type == ReqTypeValues.get("CONNECTION"):
            if Keys.get("USER_ID") in req_json:
                user = req_json[Keys.get("USER_ID")]
                req = prepare_connection_request(user)
            elif Keys.get("SESSION_ID") in req_json:
                # print("INSIDE SESSION_ID")
                session_id = req_json[Keys.get("SESSION_ID")]
                req = prepare_connection_request(session_id)
            elif Keys.get("AUTHORIZATION") in req_json:
                # print("INSIDE AUTHORIZATION")
                jwt = req_json[Keys.get("AUTHORIZATION")]
                redis_key = req_json[Keys.get("SID")]
                if jwt and redis_key:
                    req = prepareConnectionRequest2(jwt, redis_key)
                    # req = {"Authorization": jwt, "Sid": redis_key}
                else:
                    print("Authorization mode is enabled: Authorization or Sid not found !")
            else:
                print("Invalid conn mode !")
        elif req_type == ReqTypeValues.get("SCRIP_SUBS"):
            req = prepareSubsUnSubsRequest(scrips, BinRespTypes.get("SUBSCRIBE_TYPE"), SCRIP_PREFIX, channelnum)
            # print("*********** SUB SCRIPS req", req)
        elif req_type == ReqTypeValues.get("SCRIP_UNSUBS"):
            req = prepareSubsUnSubsRequest(scrips, BinRespTypes.get("UNSUBSCRIBE_TYPE"), SCRIP_PREFIX, channelnum)
        elif req_type == ReqTypeValues.get("INDEX_SUBS"):
            req = prepareSubsUnSubsRequest(scrips, BinRespTypes.get("SUBSCRIBE_TYPE"), INDEX_PREFIX, channelnum)
        elif req_type == ReqTypeValues.get("INDEX_UNSUBS"):
            req = prepareSubsUnSubsRequest(scrips, BinRespTypes.get("UNSUBSCRIBE_TYPE"), INDEX_PREFIX, channelnum)
        elif req_type == ReqTypeValues.get("DEPTH_SUBS"):
            req = prepareSubsUnSubsRequest(scrips, BinRespTypes.get("SUBSCRIBE_TYPE"), DEPTH_PREFIX, channelnum)
        elif req_type == ReqTypeValues.get("DEPTH_UNSUBS"):
            req = prepareSubsUnSubsRequest(scrips, BinRespTypes.get("UNSUBSCRIBE_TYPE"), DEPTH_PREFIX, channelnum)
        elif req_type == ReqTypeValues.get("CHANNEL_PAUSE"):
            req = prepareChannelRequest(BinRespTypes.get("CHPAUSE_TYPE"), channelnum)
        elif req_type == ReqTypeValues.get("CHANNEL_RESUME"):
            req = prepareChannelRequest(BinRespTypes.get("CHRESUME_TYPE"), channelnum)
        elif req_type == ReqTypeValues.get("SNAP_MW"):
            req = prepareSnapshotRequest(scrips, BinRespTypes.get("SNAPSHOT"), SCRIP_PREFIX)
        elif req_type == ReqTypeValues.get("SNAP_DP"):
            req = prepareSnapshotRequest(scrips, BinRespTypes.get("SNAPSHOT"), DEPTH_PREFIX)
        elif req_type == ReqTypeValues.get("SNAP_IF"):
            req = prepareSnapshotRequest(scrips, BinRespTypes.get("SNAPSHOT"), INDEX_PREFIX)
        elif req_type == ReqTypeValues.get("OPC_SUBS"):
            req = get_opc_chain_subs_request(req[Keys.get("OPC_KEY")], req[Keys.get("STK_PRC")],
                                             req[Keys.get("HIGH_STK")],
                                             req[Keys.get("LOW_STK")], channelnum)
        elif req_type == ReqTypeValues.get("THROTTLING_INTERVAL"):
            req = prepareThrottlingIntervalRequest(scrips)
        elif req_type == ReqTypeValues.get("LOG"):
            enable_log(req.get('enable'))
        if self.ws and req:
            self.ws.send(req, 0x2)
        else:
            print("Unable to send request !, Reason: Connection faulty or request not valid !")


    def is_message_for_subscription(self,message):
        # print("message ==== ",message)
        is_for_sub = False
        keys_in_sublist = list({outer_key for data_dict in self.sub_list for outer_key in data_dict})
        # print("sublist keys ",keys_in_sublist)
        for item in message:
            if 'tk' in item:
                if item['tk'] in keys_in_sublist:
                    is_for_sub = True

            if is_for_sub:
                break
        return is_for_sub

    def remove_items(self, un_sub_json):
        for unsubscribe_token in un_sub_json:
            token_value = unsubscribe_token[list(unsubscribe_token.keys())[0]]['instrument_token']
            segment_value = unsubscribe_token[list(unsubscribe_token.keys())[0]]['exchange_segment']
            sub_type_value = unsubscribe_token[list(unsubscribe_token.keys())[0]]['subscription_type']

            # self.sub_list = [token for token in self.sub_list if str(list(token.keys())[0]) != str(token_value)]
            self.sub_list = [token for token in self.sub_list if token != unsubscribe_token]

            for channel_token_list in self.channel_tokens.values():
                for channel_token_dict in channel_token_list:
                    for channel_token_key, channel_token_value in channel_token_dict.items():

                        if token_value == channel_token_value['instrument_token'] and segment_value == \
                                channel_token_value['exchange_segment'] and sub_type_value == \
                                channel_token_value['subscription_type']:
                            channel_token_list.remove(channel_token_dict)
                            break

        return

    def input_validation(self, instrument_tokens):
        valid_params = ["instrument_token", "exchange_segment"]
        ret_obj = True
        if len(instrument_tokens) > 0:
            for item in instrument_tokens:
                if ret_obj:
                    keys_lst = list(item.keys())
                    for key in valid_params:
                        if key in keys_lst:
                            pass
                        else:
                            ret_obj = False
                            break
                else:
                    break
        else:
            ret_obj = False
        return ret_obj

    def get_formatted_data(self, instrument_tokens):
        scrips = ""
        quote_type = ""
        for item in instrument_tokens:
            for k, v in item.items():
                if type(v) == dict and "exchange_segment" in v.keys() and "instrument_token" in v.keys():
                    if scrips != "":
                        scrips += '&'
                    scrips += v["exchange_segment"] + "|" + str(v["instrument_token"])
                if k == "quote_type":
                    quote_type = v
        return scrips, quote_type

    def format_tokens_live(self, instrument_tokens):
        scrips = ""
        if type(instrument_tokens) == dict and "exchange_segment" in instrument_tokens.keys() and \
                "instrument_token" in instrument_tokens.keys():
            if scrips != "":
                scrips += '&'
            scrips += instrument_tokens["exchange_segment"] + "|" + str(instrument_tokens["instrument_token"])
        return scrips

    def format_un_sub_list(self, instrument_tokens):
        scrips = ""
        for instrument_token in instrument_tokens:
            if type(instrument_token) == dict and "exchange_segment" in instrument_token.keys() and \
                    "instrument_token" in instrument_token.keys():
                if scrips != "":
                    scrips += '&'
                scrips += instrument_token["exchange_segment"] + "|" + str(instrument_token["instrument_token"])
        return scrips

    def call_quotes(self):
        scrips, quote_type = self.get_formatted_data(self.quotes_arr)
        scrip_type = ReqTypeValues.get("SNAP_MW")
        if self.quotes_index:
            scrip_type = ReqTypeValues.get("SNAP_IF")
        else:
            if quote_type:
                if quote_type.strip().lower() == 'market_depth':
                    scrip_type = ReqTypeValues.get("SNAP_DP")

        req_params = json.dumps({"type": scrip_type, "scrips": scrips, "channelnum": QuotesChannel})
        self.hs_send(req_params)

    def quote_type_validation(self, quote_type):
        Q_type = True
        if quote_type:
            if str(quote_type).strip().lower() not in ['market_depth', 'ohlc', 'ltp', '52w', 'circuit_limits',
                                                       'scrip_details']:
                Q_type = False
        return Q_type

    def get_quotes(self, instrument_tokens, quote_type=None, isIndex=None):
        if self.quote_type_validation(quote_type):
            self.quotes_index = isIndex
            # self.quotes_api_callback = callback
            if self.input_validation(instrument_tokens):
                for item in instrument_tokens:
                    key = item['instrument_token']
                    value = {'instrument_token': item['instrument_token'],
                             'exchange_segment': item['exchange_segment']}
                    if key not in [list(x.keys())[0] for x in self.quotes_arr]:
                        self.quotes_arr.append({key: value, "quote_type": quote_type})
                    else:
                        index = [list(x.keys())[0] for x in self.quotes_arr].index(key)
                        self.quotes_arr[index][key].update(value)

                self.call_quotes()

            else:
                return Exception("Invalid Inputs")
        else:
            try:
                raise ValueError(json.dumps({"Error": "Quote Type which is given is not matching",
                                             "Expected Values for quote_type": ['market_depth', 'ohlc', 'ltp',
                                                                                '52w',
                                                                                'circuit_limits',
                                                                                'scrip_details']}))
            except ValueError as e:
                print(str(e))

    def subscribe_scripts(self, channel_tokens):
        # print("self.channel_tokens.items()", self.channel_tokens)
        for channel, token_list in channel_tokens.items():
            for tokens in token_list:
                tokens = list(tokens.values())
                scrips = self.format_tokens_live(tokens[0])
                req_params1 = json.dumps(
                    {"type": tokens[0]["subscription_type"], "scrips": scrips, "channelnum": channel})
                self.hs_send(req_params1)

    def prepare_un_sub(self):
        # print("IN Prepare UNSUB")
        for key, value in self.channel_tokens.items():
            # Loop through each item in the value list
            for item in value:
                # Extract the subscription_type from the item
                subscription_type = list(item.values())[0]["subscription_type"]
                subscription_type = subscription_type.replace('s', 'u')
                # Create a new key by appending the subscription_type to the original key
                new_key = f"{key}-{subscription_type}"
                # Create a list to store the items as the value
                new_value = [{list(item.values())[0]["instrument_token"]: list(item.values())[0]}]
                # Add the item to the new_value list
                # Add the new key-value pair to un_sub_channel_token
                if new_key in self.un_sub_channel_token:
                    self.un_sub_channel_token[new_key].extend(new_value)
                else:
                    self.un_sub_channel_token[new_key] = new_value


    def get_live_feed(self, instrument_tokens, isIndex, isDepth):
        if len(self.sub_list) + len(instrument_tokens) > 3000:
            self.token_limit_reached = True
            self.prepare_un_sub()
            self.un_subscription()

        tmp_token_list = []
        subscription_type = ReqTypeValues.get("SCRIP_SUBS")
        if isIndex:
            subscription_type = ReqTypeValues.get("INDEX_SUBS")
        if isDepth:
            subscription_type = ReqTypeValues.get("DEPTH_SUBS")

        if self.input_validation(instrument_tokens):
            for item in instrument_tokens:
                key = item['instrument_token']
                value = {'instrument_token': item['instrument_token'],
                         'exchange_segment': item['exchange_segment'],
                         'subscription_type': subscription_type}
                if 'subscription_type' not in item:
                    item['subscription_type'] = subscription_type
                # if key not in [list(x.keys())[0] for x in self.sub_list]:
                if {key: value} not in self.sub_list:
                    self.sub_list.append({key: value})
                    tmp_token_list.append({key: value})
                # else:
                #     index = [list(x.keys())[0] for x in self.sub_list].index(key)
                #     print("index, key === ", index, key)
                #     print("sub list item === ", self.sub_list[index][key], item)
                #     if self.sub_list[index][key]['exchange_segment'] != item['exchange_segment'] or \
                #             self.sub_list[index][key]['subscription_type'] != item['subscription_type']:
                #         self.sub_list.append({key: value})
                #         tmp_token_list.append({key: value})
                #         print("here 2")
                #
                #     else:
                #         self.sub_list[index][key].update(value)
                #         print("here 3 ")

            channel_tokens = self.channel_segregation(tmp_token_list)
            self.subscribe_scripts(channel_tokens)
        else:
            if self.on_error:
                self.on_error(Exception("Invalid Inputs"))

    def append_ohlc_data(self, new_dict):
        new_dict["ohlc"] = {}
        if 'open' in new_dict.keys():
            new_dict["ohlc"]["open"] = new_dict['open']
            new_dict.pop('open')
        else:
            new_dict["ohlc"]["open"] = None
        if 'high' in new_dict.keys():
            new_dict["ohlc"]["high"] = new_dict['high']
            new_dict.pop('high')
        else:
            new_dict["ohlc"]["high"] = None
        if 'low' in new_dict.keys():
            new_dict["ohlc"]["low"] = new_dict['low']
            new_dict.pop('low')
        else:
            new_dict["ohlc"]["low"] = None
        if 'close' in new_dict.keys():
            new_dict["ohlc"]["close"] = new_dict['close']
            new_dict.pop('close')
        else:
            new_dict["ohlc"]["close"] = None

        return new_dict

    def quote_type_filter(self, new_dict, quote_type):
        if quote_type:
            resp_dict = {'instrument_token': new_dict['instrument_token'],
                         'trading_symbol': new_dict['trading_symbol'],
                         'exchange_segment': new_dict['exchange_segment']}
            if quote_type.strip().lower() == 'ohlc':
                resp_dict['ohlc'] = new_dict['ohlc']
                return resp_dict
            elif quote_type.strip().lower() == 'ltp':
                resp_dict['ltp'] = new_dict['last_traded_price']
                return resp_dict
            elif quote_type.strip().lower() == '52w':
                resp_dict['52week_high'] = new_dict['52week_high']
                resp_dict['52week_low'] = new_dict['52week_low']
                return resp_dict
            elif quote_type.strip().lower() == 'circuit_limits':
                resp_dict['upper_circuit_limit'] = new_dict['upper_circuit_limit']
                resp_dict['lower_circuit_limit'] = new_dict['lower_circuit_limit']
                return resp_dict
            elif quote_type.strip().lower() == 'scrip_details':
                if "open_interest" in new_dict:
                    resp_dict['open_interest'] = new_dict['open_interest']
                resp_dict['last_traded_time'] = new_dict['last_traded_time']
                resp_dict['ltp'] = new_dict['last_traded_price']
                resp_dict['last_traded_quantity'] = new_dict['last_traded_quantity']
                resp_dict['total_buy_quantity'] = new_dict['total_buy_quantity']
                resp_dict['total_sell_quantity'] = new_dict['total_sell_quantity']
                resp_dict['volume'] = new_dict['volume']
                resp_dict['average_price'] = new_dict['average_price']
                resp_dict['volume'] = new_dict['volume']
                resp_dict['change'] = new_dict['change']
                resp_dict['net_change_percentage'] = new_dict['net_change_percentage']
                return resp_dict
            else:
                return new_dict
        else:
            return new_dict

    def depth_resp_mapping(self, response_data):
        final_response = []
        for item in response_data:
            depth_resp = {
                'instrument_token': item['tk'],
                'trading_symbol': item['ts'],
                'exchange_segment': item['e'],
                'depth': {
                    'buy': [
                        {'price': item['bp'], 'quantity': item['bq'], 'orders': item['bno1']},
                        {'price': item['bp1'], 'quantity': item['bq1'], 'orders': item['bno2']},
                        {'price': item['bp2'], 'quantity': item['bq2'], 'orders': item['bno3']},
                        {'price': item['bp3'], 'quantity': item['bq3'], 'orders': item['bno4']},
                        {'price': item['bp4'], 'quantity': item['bq4'], 'orders': item['bno5']},
                    ],
                    'sell': [
                        {'price': item['sp'], 'quantity': item['bs'], 'orders': item['sno1']},
                        {'price': item['sp1'], 'quantity': item['bs1'], 'orders': item['sno2']},
                        {'price': item['sp2'], 'quantity': item['bs2'], 'orders': item['sno3']},
                        {'price': item['sp3'], 'quantity': item['bs3'], 'orders': item['sno4']},
                        {'price': item['sp4'], 'quantity': item['bs4'], 'orders': item['sno5']},
                    ]
                }
            }
            final_response.append(depth_resp)
        return final_response

    def quote_resp_mapper(self, response_data, quote_type=None):
        out_resp = []
        if len(response_data) >= 1:
            for item in response_data:
                if type(item) == dict:
                    new_dict = {stock_key_mapping.get(k, k): v for k, v in item.items()}
                    for key in list(new_dict.keys()):
                        if key not in list(stock_key_mapping.values()):
                            new_dict.pop(key)
                    new_dict = self.append_ohlc_data(new_dict)
                    if quote_type:
                        if quote_type.strip().lower() != 'market_depth':
                            out_resp.append(self.quote_type_filter(new_dict, quote_type))
                    else:
                        out_resp.append(new_dict)
                else:
                    out_resp = response_data
        return out_resp

    def quote_response_formatter(self, message):
        # print("quote response formatter ",message)
        quote_type = ''
        out_list = []
        quotes_arr_list = list(set().union(*(d.keys() for d in self.quotes_arr)))
        # print("quotes arr list ",quotes_arr_list)
        if "quote_type" in quotes_arr_list:
            quotes_arr_list.remove("quote_type")
        for item in message:
            if 'tk' in item:
                if item['tk'] in quotes_arr_list:
                    out_list.append(item)
                    for i in range(len(self.quotes_arr)):
                        if self.quotes_arr[i].get(item['tk']):
                            quote_type = self.quotes_arr[i]["quote_type"]
                            # print("quote type --- ",self.quotes_arr)
                        if self.quotes_arr[i].get(item['tk']):
                            del self.quotes_arr[i]
                            break
        return out_list, quote_type

    def response_format(self, response_data, quote_type):
        # print("response formatter ",response_data)
        # print("quote type ",quote_type)
        out_resp = []
        if self.quotes_index:
            if len(response_data) >= 1:
                for item in response_data:
                    if type(item) == dict:
                        new_dict = {index_key_mapping.get(k, k): v for k, v in item.items()}
                        for key in list(new_dict.keys()):
                            if key not in list(index_key_mapping.values()):
                                new_dict.pop(key)
                        out_resp.append(new_dict)

        else:
            if quote_type:
                if quote_type.strip().lower() == 'market_depth':
                    out_resp = self.depth_resp_mapping(response_data)
                else:
                    out_resp = self.quote_resp_mapper(response_data, quote_type)
            else:
                out_resp = self.quote_resp_mapper(response_data, quote_type)
        return out_resp

    def channel_segregation(self, tmp_token_list):
        # print("****** tmp_token_list", tmp_token_list)
        out_channel_list = {}
        for channel_num in range(2, 17):
            # Check if there is an existing channel array for this channel number
            if channel_num not in self.channel_tokens:
                self.channel_tokens[channel_num] = []
            if channel_num not in out_channel_list:
                out_channel_list[channel_num] = []

            # Check if there is space to add all the input JSON objects to this channel
            if len(self.channel_tokens[channel_num]) + len(tmp_token_list) <= 200:
                # Add all the input JSON objects to this channel
                self.channel_tokens[channel_num].extend(tmp_token_list)
                if out_channel_list[channel_num]:
                    out_channel_list[channel_num].extend(tmp_token_list)
                else:
                    out_channel_list[channel_num] = tmp_token_list
                # Exit the loop since all objects have been added to a channel
                break

            # Otherwise, add as many input JSON objects as possible to this channel
            num_to_add = 200 - len(self.channel_tokens[channel_num])
            self.channel_tokens[channel_num].extend(tmp_token_list[:num_to_add])
            out_channel_list[channel_num].extend(tmp_token_list[:num_to_add])
            # Remove the added objects from the input array
            tmp_token_list = tmp_token_list[num_to_add:]

        return out_channel_list

    def un_subscription(self):
        for channels, token_list in self.un_sub_channel_token.items():
            tokens_list = [list(tokens.values())[0] for tokens in token_list]
            scrips = self.format_un_sub_list(tokens_list)
            self.un_sub_channel = channels
            channel, sub_type = channels.split('-')
            req_params1 = json.dumps(
                {"type": sub_type, "scrips": scrips, "channelnum": channel})
            self.hs_send(req_params1)

    def un_subscribe_list(self, instrument_tokens, isIndex=False, isDepth=False):
        # print("INTO UNSUBSCRIBE", instrument_tokens)
        un_subscription_type = ReqTypeValues.get("SCRIP_UNSUBS")
        subscription_type = ReqTypeValues.get("SCRIP_SUBS")
        if isIndex:
            un_subscription_type = ReqTypeValues.get("INDEX_UNSUBS")
            subscription_type = ReqTypeValues.get("INDEX_SUBS")
        if isDepth:
            un_subscription_type = ReqTypeValues.get("DEPTH_UNSUBS")
            subscription_type = ReqTypeValues.get("DEPTH_SUBS")

        if self.input_validation(instrument_tokens):
            extracted_tokens = [{'instrument_token': item[key]['instrument_token'],
                                 'exchange_segment': item[key]['exchange_segment'],
                                 'subscription_type': item[key]['subscription_type']}
                                for item in self.sub_list for key in item]

            for token in instrument_tokens:
                token["subscription_type"] = subscription_type
                if token in extracted_tokens:
                    for key, value in self.channel_tokens.items():
                        for obj in value:
                            if list(obj.values())[0] == token:
                                in_key = token['instrument_token']
                                value = {'instrument_token': token['instrument_token'],
                                         'exchange_segment': token['exchange_segment'],
                                         'subscription_type': subscription_type}
                                key = str(key) + '-' + un_subscription_type
                                if key not in self.un_sub_channel_token:
                                    self.un_sub_channel_token[key] = []
                                self.un_sub_channel_token[key].append({in_key: value})

                else:
                    print("The Given Token is not in Subscription list")
            if self.ws:
                self.un_subscription()

            else:
                print("Socket Connection has been closed, So! The scripts are already un-subscribed!")


if __name__ == '__main__':
    token = "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJzY29wZSI6WyJWaWV3Il0sImV4cCI6MTcxOTUxMzAwMCwianRpIjoiYzBlMTA0ZWMtMjYxMi00YTBjLWFlNWYtNjNiN2QyZWE3NDZjIiwiaWF0IjoxNzE5NDcyOTE4LCJpc3MiOiJsb2dpbi1zZXJ2aWNlIiwic3ViIjoiYThkODA0MTYtOTRjYi00Nzg0LTg0YjctYTRkZDExYTgzMDA2IiwidWNjIjoiWjZZQzAiLCJuYXAiOiIiLCJ5Y2UiOiJlWVxcNiQnOTVcInNcdFxuXHUwMDAwclx1MDAwMFx1MDAxMGIiLCJmZXRjaGNhY2hpbmdydWxlIjowLCJjYXRlZ29yaXNhdGlvbiI6IiJ9.qUoHAL2Duor6DfZ_wa6S_T-95CRWACp0ubUlBegrL16f1eYzPyIjVyuff-5vAmc4RxQyoYlQ-Ie_tK_uXEgD7CDg1klxW32IsYLg-EPtGaxaqWbXgBmu9stoxCAtodgNa9Umm3KwXWRTtsMIlRbXn0QiiATmRsbmrcH_UESNy4CKpVAf3nS85ymPB3R1KPC9dsI5ZJaPSpP_O_t-J1TT0czaGH7rotgQ39cnTnl4YEs07DluCryfra7maDfzRnZ2NxiQhUhEALwZnHuRdj2Ihx6YJXEoOpMsY_y41TvkeB5enm2ApN2XcnGuV6ny5BR_KjemxZvAlfGVBki5bqZDSQ"
    sid = "c8f7cde9-e44f-4df4-9fc9-77997cc9c2e1"
    hsServerId = "server3"

    def custom_on_open():
        inst_tokens = [{"instrument_token": "11536", "exchange_segment": "nse_cm"},
                       {"instrument_token": "1594", "exchange_segment": "nse_cm"},
                       {"instrument_token": "11915", "exchange_segment": "nse_cm"},
                       {"instrument_token": "13245", "exchange_segment": "nse_cm"}]
        soc.get_live_feed(inst_tokens, isIndex=False, isDepth=False)
        print("Custom open handler")

    def custom_on_message(message):
        print(f"Custom message handler: {message}")

    def custom_on_close(close_status_code, close_msg):
        print("Custom close handler : ", close_status_code, close_msg)

    def custom_on_error(error):
        print(f"Custom error handler: {error}")


    soc = KotakDataFeed(token,sid,hsServerId,custom_on_message,custom_on_error,custom_on_close,custom_on_open,retry_attempts=2)
    soc.start_websocket()
    time.sleep(2)
    soc.close_websocket()





from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
try: 
    long_description = (this_directory / "readme.md").read_text()
except:
    long_description = "Official Library for spark.maticalgos.com"

def parse_requirements(filename):
    with open(filename, 'r') as f:
        return f.read().splitlines()
    
setup(
name='maticalgos-spark',
version='0.2.8',
author='Niraj Munot',
author_email='nirajmunot28@gmail.com',
description='SparkLib is a Python client library for interacting with https://spark.maticalgos.com . It provides functionalities to manage accounts, strategies, place orders, and much more.',
long_description=long_description,
include_package_data=True,
package_data={"map.json":['maticalgos/sparkLib/utility/dataws/dependencies/fyersAPIv3/map.json']},
data_files=[("map.json", ['maticalgos/sparkLib/utility/dataws/dependencies/fyersAPIv3/map.json'])],
long_description_content_type='text/markdown',
packages=find_packages(),
install_requires=['websocket-client==1.7.0', 'numpy==1.26.4', 'requests==2.31.0'],
extras_require={'dataws' : ['python-engineio==3.13.0', 'python-socketio==4.6.0', 'aws-lambda-powertools==1.25.5', 'websockets==13.0.1', 'protobuf==5.28.1'], 
                'wsHandler' : ['python-engineio==3.13.0', 'python-socketio==4.6.0','aws-lambda-powertools==1.25.5', 
                               'websockets==13.0.1', 'protobuf==5.28.1',
                               'pyzmq==26.2.0', 'redis==4.3.4'],
                "full" : ['python-engineio==3.13.0', 'python-socketio==4.6.0','aws-lambda-powertools==1.25.5', 
                               'websockets==13.0.1', 'protobuf==5.28.1','websocket-client==1.7.0', 'numpy==1.26.4', 'requests==2.31.0', #websockets==8.1 protobuf==3.20.1
                               'pyzmq==26.2.0', 'redis==4.3.4','duckdb==1.0.0','datetime==4.3', 'pandas==2.2.2']}, ##pyzmq==26.2.0 #old 23.2.0
classifiers=[
'Programming Language :: Python :: 3',
'License :: OSI Approved :: MIT License',
'Operating System :: OS Independent',
],
python_requires='>=3.6',
)


import datetime
import traceback
import os
from .dependencies import SmartWebSocketV2, SparkLib

class AngelWs():
    correlation_id = "abcd1234"
    BROKER = "ANGELONE"
    TOKENS = {}
    def __init__(self, accountData, accessToken= None, dataType = "ltp", 
                 onLtp = None, 
                 onDepth = None, 
                 onError = None, 
                 onClose = None,
                 onOpen = None,
                 ) :
        self.accountData = accountData
        self.dType = 3 if dataType == "ltp" else 3 if dataType == "DepthUpdate" else None #Currently Only supports LTP Updates.
        self.onLtp = self.__onLTP if not onLtp  else onLtp
        self.onDepth = self.__onDepth if not onDepth else onDepth
        self.onError = self.__onError if not onError else onError
        self.onClose = self.__onClose if not onClose else onClose
        self.onOpen = self.__onOpen if not onOpen else onOpen
        self.run = True
        if not accessToken :
            if os.environ.get("MATICALGOS_AccessToken"):
                accessToken = os.environ["MATICALGOS_AccessToken"] 
            else: 
                raise Exception("Please generate access token.")
        self.spk = SparkLib(access_token=accessToken)
    
    def __onLTP(self, message):
        print(message)
    
    def __onDepth(self, message):
        print(message)
    
    def __onError(self, message):
        print("error : ",message)

    def __onClose(self, *args):
        print("Connection Closed")

    def __onOpen(self):
        print("Connection Opened")

    def getTokens(self, tokens):
        data = self.spk.getBrokerTokens(tokens=tokens, broker=self.BROKER)
        if data['status'] and not data['error'] : 
            tokens = data['data']
            return tokens
        else: 
            raise Exception (data['message'])
        
    def getSymbol(self, exchange, token):
        return self.TOKENS[", ".join([str(exchange), str(token)])]
    
    def tickHandler(self, message):
        try: 
            token = self.getSymbol(message['exchange_type'], message['token'])
            ts = int(message['exchange_timestamp']/1000) #message['last_traded_timestamp'] if str(message['last_traded_timestamp']) != "0" else 
            msg = {"timestamp_str" : str(datetime.datetime.fromtimestamp(ts)),
                "timestamp" : str(ts),
                "symbol" : token, 
                "ltp" : float(message['last_traded_price']/100), 
                "prev_day_close" : float(message['closed_price']/100), 
                "oi" : int(message['open_interest']), 
                "prev_day_oi" : 0, 
                "turnover" : int(message['volume_trade_for_the_day']) , 
                "best_bid_price" : float(message['best_5_buy_data'][0]['price']/100) if message.get('best_5_buy_data') else 0, 
                "best_ask_price" : float(message['best_5_sell_data'][-1]['price']/100) if message.get('best_5_sell_data') else 0, 
                "best_bid_qty" : int(message['best_5_buy_data'][0]['quantity']) if message.get('best_5_buy_data') else 0,
                "best_ask_qty" : int(message['best_5_sell_data'][-1]['quantity']) if message.get('best_5_sell_data') else 0, 
                "ttq" : int(message['volume_trade_for_the_day']),  
                "token" : token}
            self.onLtp(msg)
        except Exception as e : 
            self.onError(f"Error : {e}, Traceback : {traceback.format_exc()}")

    def disconnect(self):
        self.run = False
        self.ang.close_connection()
    
    def Subscribe(self, tokens):
        try: 
            tokens = self.getTokens(tokens)
            tkManage = {}
            notFoundTK = []
            
            for tk in tokens:
                wsToken = tk['wsToken'] 
                if wsToken != None : 
                    exch, tok = wsToken.split(", ")
                    if not tkManage.get(exch) : 
                        tkManage[exch] = {"action" : 1, "tokens" : [], "exchangeType" : exch}
                    tkManage[exch]['tokens'].append(tok)
                    self.TOKENS[wsToken] = tk['token']
                else: 
                    notFoundTK.append(tk['token'])

            lst = [tkManage[i] for i in tkManage]
            self.ang.subscribe(correlation_id=self.correlation_id, mode=self.dType, token_list=lst)
            if notFoundTK != [] : 
                self.onError(f"ERROR : Unable to find tokens for {notFoundTK}")

        except Exception as e :
            self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")

    def Unsubscribe(self, tokens):
        tkManage = {}
        try: 
            tokens = self.getTokens(tokens)
            for tk in tokens: 
                wsToken = tk['wsToken']
                if wsToken != None : 
                    exch, tok = wsToken.split(", ")
                    if not tkManage.get(exch) : 
                        tkManage[exch] = {"action" : 0, "tokens" : [], "exchangeType" : int(exch)}
                    tkManage[exch]['tokens'].append(tok[3:])

            lst = [tkManage[i] for i in tkManage]
            self.ang.unsubscribe(correlation_id=self.correlation_id, mode=self.dType, token_list=lst)

        except Exception as e :
            self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")
            
    def connect(self):
        try: 
            sessionData = self.accountData['Sessionid']
            self.ang = SmartWebSocketV2(auth_token=sessionData['jwtToken'],
                                        api_key=self.accountData['ApiKey'],
                                        client_code=self.accountData['Clientid'],
                                        feed_token=sessionData['feedToken'],
                                        max_retry_attempt=5,
                                        on_Message=self.tickHandler,
                                        on_Error=self.onError,
                                        on_Close=self.onClose,
                                        on_Open=self.onOpen)
            
            self.ang.connect()
        except Exception as e : 
            self.onClose(f"Error : {e}, Traceback : {traceback.format_exc()}")



import datetime
import traceback
import os
from .dependencies import ShoonyaDataWS, SparkLib

class ShoonyaWS():
    dataStore = {}
    BROKER = "SHOONYA"
    TOKENS = {}
    def __init__(self, accountData, accessToken = None, dataType = "ltp", 
                 onLtp = None, 
                 onDepth = None, 
                 onError = None, 
                 onClose = None,
                 onOpen = None,
                 ) :
        self.accountData = accountData
        self.dType = 1 if dataType == "ltp" else 2 if dataType == "DepthUpdate" else None #Currently Only supports LTP Updates.
        self.onLtp = self.__onLTP if not onLtp  else onLtp
        self.onDepth = self.__onDepth if not onDepth else onDepth
        self.onError = self.__onError if not onError else onError
        self.onClose = self.__onClose if not onClose else onClose
        self.onOpen = self.__onOpen if not onOpen else onOpen
        self.run = True
        if not accessToken :
            if os.environ.get("MATICALGOS_AccessToken"):
                accessToken = os.environ["MATICALGOS_AccessToken"] 
            else: 
                raise Exception("Please generate access token.")
        self.spk = SparkLib(access_token=accessToken)
    
    def __onLTP(self, message):
        print(message)
    
    def __onDepth(self, message):
        print(message)
    
    def __onError(self, message):
        print("error : ",message)

    def __onClose(self, *args):
        print("Connection Closed")

    def __onOpen(self):
        print("Connection Opened")
    
    def getTokens(self, tokens):
        data = self.spk.getBrokerTokens(tokens=tokens, broker=self.BROKER)
        if data['status'] and not data['error'] : 
            tokens = data['data']
            return tokens
        else: 
            raise Exception (data['message'])
    
    def getSymbol(self, exchange, token):
        return self.TOKENS["|".join([exchange, token])]

    def tickHandler(self, message):
        if message['t'] in ['tk', 'tf'] : 
            try:
                data = {
                        "time" : message.get('ft'),
                        "bidPrice" : message.get('bp1') if message.get('bp1') != None else message.get('lp') if message.get('lp') != None else 0 ,
                        "askPrice" : message.get('sp') if message.get('sp') != None else message.get('lp') if message.get('lp') != None else 0,
                        "bidQty" : message.get('bq1') if message.get('bq1') != None else 0,
                        "askQty" : message.get('sq1') if message.get('sq1') != None else 0,
                        "token" : message.get('tk'),
                        "exchange" : message.get('e'), 
                        "ltp" : message.get('lp') if message.get('lp') != None else (float(message['bp1'])+float(message['sp1']))/2 if None not in [message.get('bp1'),message.get('sp1')] else None
                        }
                if None not in data.values() : 
                    token = self.getSymbol(message['e'], message['tk'])
                    dtime = datetime.datetime.fromtimestamp(int(data['time']))
                    openInterest = 0 if token.split(":")[0] not in ['NFO',"BFO","MCX"] else message.get('oi') 
                    if openInterest == None :
                        openInterest = self.dataStore[token]['oi'] if self.dataStore.get(token) != None and self.dataStore[token].get('v') else 0
                    volume = message.get('v') if (message.get('v') != None and message.get('v') != 0) \
                                else self.dataStore[token]['v'] if (self.dataStore.get(token) != None and self.dataStore[token].get('v')) else 0 
                    volume = 0 if volume == None else volume
                    turnover = volume
                    prevClose = message.get('c') if message.get('c') != None \
                                else round(float(message.get('lp'))/(1-(float(message.get('pc'))/100)),2) if (message.get('pc') != None and message.get('lp') != None) \
                                else self.dataStore.get(token)['c'] if self.dataStore.get(token) != None and self.dataStore[token].get('c') != None else \
                                0.0
                              
                    # prevClose = None
                    # if self.dataStore.get(token) != None :
                    #     prevClose = self.dataStore[token]['c'] if self.dataStore[token].get("c") != None else None
                        
                    # if not self.dataStore.get(token) or prevClose == None :  
                    #     prevClose = message.get('c') if message.get('c') != None else \
                    #             float(data['ltp'])/(1-(float(message.get('pc'))/100)) if message.get('pc') != None \
                    #             else 0.0
                                    
                    msg = {
                                    "timestamp_str" : str(dtime),
                                    "timestamp" : data['time'],
                                    "symbol" : token, 
                                    "ltp" : float(data['ltp']), 
                                    "prev_day_close" : prevClose, 
                                    "oi" : int(openInterest), 
                                    "prev_day_oi" : 0, 
                                    "turnover" : int(turnover) , 
                                    "best_bid_price" : float(data['bidPrice']), 
                                    "best_ask_price" : float(data['askPrice']), 
                                    "best_bid_qty" : int(data["bidQty"]),
                                    "best_ask_qty" : int(data['askQty']),
                                    "ttq" : int(volume),  
                                    "token" : token
                                    }
                    self.onLtp(msg)
                    self.dataStore[token] = {"c" : prevClose, 'oi' : openInterest, 'to' : turnover, "v" : volume}
                else: 
                    if None not in  [message.get('tk'),message.get('e')]:
                        token = self.getSymbol(message.get('e'), message.get('tk'))
                        prevclose = message.get('c') if message.get('c') != None \
                                    else float(message.get('lp'))/(1-(float(message.get('pc'))/100)) if (message.get('pc') != None and message.get('lp') != None) \
                                    else None
                        data = {
                                "c" : prevclose,
                                "oi" : message.get('oi'),
                                "v" : message.get("v"),
                                'to' : message.get('v')
                            }
                        if self.dataStore.get(token) != None : 
                            data = {k : data[k] for k in data if data[k] != None }
                            self.dataStore[token].update(data)
                        else: 
                            data = {k : 0 for k in data if data[k] == None }
                            self.dataStore[token] = data
                            
            except Exception as e : 
                self.onError(f"Error : {e}, data : {message}, Traceback : {traceback.format_exc()}")

    def disconnect(self):
        self.run = False
        self.soc.close_websocket()
    
    def Subscribe(self, tokens):
        try: 
            tokens = self.getTokens(tokens)
            toSubs = []
            notFound = []
            for tk in tokens:
                wsToken = tk['wsToken'] 
                if wsToken != None : 
                    self.TOKENS[wsToken] = tk['token']
                    toSubs.append(wsToken)
                else: 
                    notFound.append(tk['token'])
            if toSubs != [] : 
                self.soc.subscribe(instrument=toSubs, feed_type=self.dType)
            if notFound != [] : 
                self.onError(f"Error : Unable to find tokens for {notFound} ")
        except Exception as e :
            self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")

    def Unsubscribe(self, tokens):
        try: 
            tokens = self.getTokens(tokens)
            toSubs = []
            notFound = []
            for tk in tokens:
                wsToken = tk['wsToken'] 
                if wsToken != None : 
                    toSubs.append(wsToken)
                else: 
                    notFound.append(tk['token'])
            if toSubs != [] : 
                self.soc.unsubscribe(toSubs, feed_type=self.dType)
            if notFound != [] : 
                self.onError(f"Error : Unable to find tokens for {notFound} ")
        except Exception as e :
            self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")
            
    def connect(self):
        try: 
            sessionData = self.accountData['Sessionid']
            self.soc = ShoonyaDataWS(message_update_callback=self.tickHandler,
                   socket_open_callback=self.onOpen,
                   socket_close_callback=self.onClose,
                   socket_error_callback=self.onError,
                   susertoken=sessionData,
                   userid=self.accountData['Clientid'],
                   retry_attempts=4)
            self.soc.start_websocket()

        except Exception as e : 
            self.onClose(f"Error : {e}, Traceback : {traceback.format_exc()}")
